from __future__ import annotations

import gc
import operator
import platform
import sys
import weakref
from collections import deque
from collections.abc import Callable, Iterable, Iterator, KeysView, Mapping
from types import ModuleType
from typing import TypeVar, Union, cast

import pytest

import multidict
from multidict import (
    CIMultiDict,
    MultiDict,
    MultiDictProxy,
    MultiMapping,
    MutableMultiMapping,
    istr,
)

_T = TypeVar("_T")
IS_PYPY = platform.python_implementation() == "PyPy"


def chained_callable(
    module: ModuleType,
    callables: Iterable[str],
) -> Callable[..., MultiMapping[int | str] | MutableMultiMapping[int | str]]:
    """
    Return callable that will get and call all given objects in module in
    exact order.
    """

    def chained_call(
        *args: object,
        **kwargs: object,
    ) -> MultiMapping[int | str] | MutableMultiMapping[int | str]:
        callable_chain = (getattr(module, name) for name in callables)
        first_callable = next(callable_chain)

        value = first_callable(*args, **kwargs)
        for element in callable_chain:
            value = element(value)

        return cast(
            Union[
                MultiMapping[Union[int, str]],
                MutableMultiMapping[Union[int, str]],
            ],
            value,
        )

    return chained_call


@pytest.fixture
def cls(  # type: ignore[misc]
    request: pytest.FixtureRequest,
    multidict_module: ModuleType,
) -> Callable[..., MultiMapping[int | str] | MutableMultiMapping[int | str]]:
    """Make a callable from multidict module, requested by name."""
    return chained_callable(multidict_module, request.param)


def test_exposed_names(any_multidict_class_name: str) -> None:
    assert any_multidict_class_name in multidict.__all__


@pytest.mark.parametrize(
    ("cls", "key_cls"),
    (
        (("MultiDict",), str),
        (
            ("MultiDict", "MultiDictProxy"),
            str,
        ),
    ),
    indirect=["cls"],
)
def test__iter__types(
    cls: type[MultiDict[Union[str, int]]],
    key_cls: type[str],
) -> None:
    d = cls([("key", "one"), ("key2", "two"), ("key", 3)])
    for i in d:
        assert type(i) is key_cls, (type(i), key_cls)


def test_proxy_copy(
    any_multidict_class: type[MultiDict[str]],
    any_multidict_proxy_class: type[MultiDictProxy[str]],
) -> None:
    d1 = any_multidict_class(key="value", a="b")
    p1 = any_multidict_proxy_class(d1)

    d2 = p1.copy()
    assert d1 == d2
    assert d1 is not d2


def test_multidict_subclassing(
    any_multidict_class: type[MultiDict[str]],
) -> None:
    class DummyMultidict(any_multidict_class):  # type: ignore[valid-type,misc]
        pass


def test_multidict_proxy_subclassing(
    any_multidict_proxy_class: type[MultiDictProxy[str]],
) -> None:
    class DummyMultidictProxy(
        any_multidict_proxy_class,  # type: ignore[valid-type,misc]
    ):
        pass


class BaseMultiDictTest:
    def test_instantiate__empty(self, cls: type[MutableMultiMapping[str]]) -> None:
        d = cls()
        empty: Mapping[str, str] = {}
        assert d == empty
        assert len(d) == 0
        assert list(d.keys()) == []
        assert list(d.values()) == []
        assert list(d.items()) == []

        assert cls() != list()  # type: ignore[comparison-overlap]
        with pytest.raises(TypeError, match=r"3 were given"):
            cls(("key1", "value1"), ("key2", "value2"))  # type: ignore[call-arg]  # noqa: E501

    @pytest.mark.parametrize("arg0", ([("key", "value1")], {"key": "value1"}))
    def test_instantiate__from_arg0(
        self,
        cls: type[MultiDict[str]],
        arg0: Union[list[tuple[str, str]], dict[str, str]],
    ) -> None:
        d = cls(arg0)

        assert d == {"key": "value1"}
        assert len(d) == 1
        assert list(d.keys()) == ["key"]
        assert list(d.values()) == ["value1"]
        assert list(d.items()) == [("key", "value1")]

    def test_instantiate__with_kwargs(
        self,
        cls: type[MultiDict[str]],
    ) -> None:
        d = cls([("key", "value1")], key2="value2")

        assert d == {"key": "value1", "key2": "value2"}
        assert len(d) == 2
        assert sorted(d.keys()) == ["key", "key2"]
        assert sorted(d.values()) == ["value1", "value2"]
        assert sorted(d.items()) == [("key", "value1"), ("key2", "value2")]

    def test_instantiate__from_generator(
        self, cls: Union[type[MultiDict[int]], type[CIMultiDict[int]]]
    ) -> None:
        d = cls((str(i), i) for i in range(2))

        assert d == {"0": 0, "1": 1}
        assert len(d) == 2
        assert sorted(d.keys()) == ["0", "1"]
        assert sorted(d.values()) == [0, 1]
        assert sorted(d.items()) == [("0", 0), ("1", 1)]

    def test_instantiate__from_list_of_lists(
        self,
        cls: type[MutableMultiMapping[str]],
    ) -> None:
        # Should work at runtime, but won't type check.
        d = cls([["key", "value1"]])  # type: ignore[call-arg]
        assert d == {"key": "value1"}

    def test_instantiate__from_list_of_custom_pairs(
        self,
        cls: type[MultiDict[str]],
    ) -> None:
        class Pair:
            def __len__(self) -> int:
                return 2

            def __getitem__(self, pos: int) -> str:
                return ("key", "value1")[pos]

        # Works at runtime, but won't type check.
        d = cls([Pair()])  # type: ignore[list-item]
        assert d == {"key": "value1"}

    def test_getone(self, cls: type[MultiDict[str]]) -> None:
        d = cls([("key", "value1")], key="value2")

        assert d.getone("key") == "value1"
        assert d.get("key") == "value1"
        assert d["key"] == "value1"

        with pytest.raises(KeyError, match="key2"):
            d["key2"]
        with pytest.raises(KeyError, match="key2"):
            d.getone("key2")

        assert d.getone("key2", "default") == "default"

    def test_call_with_kwargs(self, cls: type[MultiDict[str]]) -> None:
        d = cls([("present", "value")])
        assert d.getall(default="missing", key="notfound") == "missing"

    def test__iter__(
        self,
        cls: Union[
            type[MultiDict[Union[str, int]]],
            type[CIMultiDict[Union[str, int]]],
        ],
    ) -> None:
        d = cls([("key", "one"), ("key2", "two"), ("key", 3)])
        assert list(d) == ["key", "key2", "key"]

    def test__contains(
        self,
        cls: Union[
            type[MultiDict[Union[str, int]]],
            type[CIMultiDict[Union[str, int]]],
        ],
    ) -> None:
        d = cls([("key", "one"), ("key2", "two"), ("key", 3)])

        assert list(d) == ["key", "key2", "key"]

        assert "key" in d
        assert "key2" in d

        assert "foo" not in d
        assert 42 not in d  # type: ignore[comparison-overlap]

    def test_keys__contains(
        self,
        cls: Union[
            type[MultiDict[Union[str, int]]],
            type[CIMultiDict[Union[str, int]]],
        ],
    ) -> None:
        d = cls([("key", "one"), ("key2", "two"), ("key", 3)])

        assert list(d.keys()) == ["key", "key2", "key"]

        assert "key" in d.keys()
        assert "key2" in d.keys()

        assert "foo" not in d.keys()
        assert 42 not in d.keys()  # type: ignore[comparison-overlap]

    def test_values__contains(
        self,
        cls: Union[
            type[MultiDict[Union[str, int]]],
            type[CIMultiDict[Union[str, int]]],
        ],
    ) -> None:
        d = cls([("key", "one"), ("key", "two"), ("key", 3)])

        assert list(d.values()) == ["one", "two", 3]

        assert "one" in d.values()
        assert "two" in d.values()
        assert 3 in d.values()

        assert "foo" not in d.values()

    def test_items__contains(
        self,
        cls: Union[
            type[MultiDict[Union[str, int]]],
            type[CIMultiDict[Union[str, int]]],
        ],
    ) -> None:
        d = cls([("key", "one"), ("key", "two"), ("key", 3)])

        assert list(d.items()) == [("key", "one"), ("key", "two"), ("key", 3)]

        assert ("key", "one") in d.items()
        assert ("key", "two") in d.items()
        assert ("key", 3) in d.items()

        assert ("foo", "bar") not in d.items()
        assert (42, 3) not in d.items()  # type: ignore[comparison-overlap]
        assert 42 not in d.items()  # type: ignore[comparison-overlap]

    def test_cannot_create_from_unaccepted(
        self,
        cls: type[MutableMultiMapping[str]],
    ) -> None:
        with pytest.raises(ValueError, match="multidict update sequence element"):
            cls([(1, 2, 3)])  # type: ignore[call-arg]

    def test_keys_is_set_less(self, cls: type[MultiDict[str]]) -> None:
        d = cls([("key", "value1")])

        assert d.keys() < {"key", "key2"}

    @pytest.mark.parametrize(
        ("contents", "expected"),
        (
            ([("key", "value1")], True),
            ([("key", "value1"), ("key2", "value2")], True),
            ([("key", "value1"), ("key2", "value2"), ("key3", "value3")], False),
            ([("key", "value1"), ("key3", "value3")], False),
        ),
    )
    def test_keys_is_set_less_equal(
        self,
        cls: type[MultiDict[str]],
        contents: list[tuple[str, str]],
        expected: bool,
    ) -> None:
        d = cls(contents)

        result = d.keys() <= {"key", "key2"}
        assert result is expected

    def test_keys_is_set_equal(self, cls: type[MultiDict[str]]) -> None:
        d = cls([("key", "value1")])

        assert d.keys() == {"key"}

    def test_items_is_set_equal(self, cls: type[MultiDict[str]]) -> None:
        d = cls([("key", "value1")])

        assert d.items() == {("key", "value1")}

    def test_keys_is_set_greater(self, cls: type[MultiDict[str]]) -> None:
        d = cls([("key", "value1"), ("key2", "value2")])

        assert d.keys() > {"key"}

    @pytest.mark.parametrize(
        ("set_", "expected"),
        (
            ({"key"}, True),
            ({"key", "key2"}, True),
            ({"key", "key2", "key3"}, False),
            ({"key3"}, False),
        ),
    )
    def test_keys_is_set_greater_equal(
        self, cls: type[MultiDict[str]], set_: set[str], expected: bool
    ) -> None:
        d = cls([("key", "value1"), ("key2", "value2")])

        result = d.keys() >= set_
        assert result is expected

    def test_keys_less_than_not_implemented(self, cls: type[MultiDict[str]]) -> None:
        d = cls([("key", "value1")])

        sentinel_operation_result = object()

        class RightOperand:
            def __gt__(self, other: KeysView[str]) -> object:
                assert isinstance(other, KeysView)
                return sentinel_operation_result

        assert (d.keys() < RightOperand()) is sentinel_operation_result

    def test_keys_less_than_or_equal_not_implemented(
        self, cls: type[MultiDict[str]]
    ) -> None:
        d = cls([("key", "value1")])

        sentinel_operation_result = object()

        class RightOperand:
            def __ge__(self, other: KeysView[str]) -> object:
                assert isinstance(other, KeysView)
                return sentinel_operation_result

        assert (d.keys() <= RightOperand()) is sentinel_operation_result

    def test_keys_greater_than_not_implemented(self, cls: type[MultiDict[str]]) -> None:
        d = cls([("key", "value1")])

        sentinel_operation_result = object()

        class RightOperand:
            def __lt__(self, other: KeysView[str]) -> object:
                assert isinstance(other, KeysView)
                return sentinel_operation_result

        assert (d.keys() > RightOperand()) is sentinel_operation_result

    def test_keys_greater_than_or_equal_not_implemented(
        self, cls: type[MultiDict[str]]
    ) -> None:
        d = cls([("key", "value1")])

        sentinel_operation_result = object()

        class RightOperand:
            def __le__(self, other: KeysView[str]) -> object:
                assert isinstance(other, KeysView)
                return sentinel_operation_result

        assert (d.keys() >= RightOperand()) is sentinel_operation_result

    def test_keys_is_set_not_equal(self, cls: type[MultiDict[str]]) -> None:
        d = cls([("key", "value1")])

        assert d.keys() != {"key2"}

    def test_keys_not_equal_unrelated_type(self, cls: type[MultiDict[str]]) -> None:
        d = cls([("key", "value1")])

        assert d.keys() != "other"  # type: ignore[comparison-overlap]

    def test_eq(self, cls: type[MultiDict[str]]) -> None:
        d = cls([("key", "value1")])

        assert {"key": "value1"} == d

    def test_eq2(self, cls: type[MultiDict[str]]) -> None:
        d1 = cls([("key", "value1")])
        d2 = cls([("key2", "value1")])

        assert d1 != d2

    def test_eq3(self, cls: type[MultiDict[str]]) -> None:
        d1 = cls([("key", "value1")])
        d2 = cls()

        assert d1 != d2

    def test_eq_other_mapping_contains_more_keys(
        self,
        cls: type[MultiDict[str]],
    ) -> None:
        d1 = cls(foo="bar")
        d2 = dict(foo="bar", bar="baz")

        assert d1 != d2

    def test_eq_bad_mapping_len(
        self, cls: Union[type[MultiDict[int]], type[CIMultiDict[int]]]
    ) -> None:
        class BadMapping(Mapping[str, int]):
            def __getitem__(self, key: str) -> int:
                return 1  # pragma: no cover  # `len()` fails earlier

            def __iter__(self) -> Iterator[str]:
                yield "a"  # pragma: no cover  # `len()` fails earlier

            def __len__(self) -> int:
                return 1 // 0

        d1 = cls(a=1)
        d2 = BadMapping()
        with pytest.raises(ZeroDivisionError):
            d1 == d2

    def test_eq_bad_mapping_getitem(
        self,
        cls: Union[type[MultiDict[int]], type[CIMultiDict[int]]],
    ) -> None:
        class BadMapping(Mapping[str, int]):
            def __getitem__(self, key: str) -> int:
                return 1 // 0

            def __iter__(self) -> Iterator[str]:
                yield "a"  # pragma: no cover  # foreign objects no iterated

            def __len__(self) -> int:
                return 1

        d1 = cls(a=1)
        d2 = BadMapping()
        with pytest.raises(ZeroDivisionError):
            d1 == d2

    def test_ne(self, cls: type[MultiDict[str]]) -> None:
        d = cls([("key", "value1")])

        assert d != {"key": "another_value"}

    def test_and(self, cls: type[MultiDict[str]]) -> None:
        d = cls([("key", "value1")])

        assert {"key"} == d.keys() & {"key", "key2"}

    def test_and2(self, cls: type[MultiDict[str]]) -> None:
        d = cls([("key", "value1")])

        assert {"key"} == {"key", "key2"} & d.keys()

    def test_bitwise_and_not_implemented(self, cls: type[MultiDict[str]]) -> None:
        d = cls([("key", "value1")])

        sentinel_operation_result = object()

        class RightOperand:
            def __rand__(self, other: KeysView[str]) -> object:
                assert isinstance(other, KeysView)
                return sentinel_operation_result

        assert d.keys() & RightOperand() is sentinel_operation_result

    def test_bitwise_and_iterable_not_set(self, cls: type[MultiDict[str]]) -> None:
        d = cls([("key", "value1")])

        assert {"key"} == d.keys() & ["key", "key2"]

    def test_or(self, cls: type[MultiDict[str]]) -> None:
        d = cls([("key", "value1")])

        assert {"key", "key2"} == d.keys() | {"key2"}

    def test_or2(self, cls: type[MultiDict[str]]) -> None:
        d = cls([("key", "value1")])

        assert {"key", "key2"} == {"key2"} | d.keys()

    def test_bitwise_or_not_implemented(self, cls: type[MultiDict[str]]) -> None:
        d = cls([("key", "value1")])

        sentinel_operation_result = object()

        class RightOperand:
            def __ror__(self, other: KeysView[str]) -> object:
                assert isinstance(other, KeysView)
                return sentinel_operation_result

        assert d.keys() | RightOperand() is sentinel_operation_result

    def test_bitwise_or_iterable_not_set(self, cls: type[MultiDict[str]]) -> None:
        d = cls([("key", "value1")])

        assert {"key", "key2"} == d.keys() | ["key2"]

    def test_sub(self, cls: type[MultiDict[str]]) -> None:
        d = cls([("key", "value1"), ("key2", "value2")])

        assert {"key"} == d.keys() - {"key2"}

    def test_sub2(self, cls: type[MultiDict[str]]) -> None:
        d = cls([("key", "value1"), ("key2", "value2")])

        assert {"key3"} == {"key", "key2", "key3"} - d.keys()

    def test_sub_not_implemented(self, cls: type[MultiDict[str]]) -> None:
        d = cls([("key", "value1"), ("key2", "value2")])

        sentinel_operation_result = object()

        class RightOperand:
            def __rsub__(self, other: KeysView[str]) -> object:
                assert isinstance(other, KeysView)
                return sentinel_operation_result

        assert d.keys() - RightOperand() is sentinel_operation_result

    def test_sub_iterable_not_set(self, cls: type[MultiDict[str]]) -> None:
        d = cls([("key", "value1"), ("key2", "value2")])

        assert {"key"} == d.keys() - ["key2"]

    def test_xor(self, cls: type[MultiDict[str]]) -> None:
        d = cls([("key", "value1"), ("key2", "value2")])

        assert {"key", "key3"} == d.keys() ^ {"key2", "key3"}

    def test_xor2(self, cls: type[MultiDict[str]]) -> None:
        d = cls([("key", "value1"), ("key2", "value2")])

        assert {"key", "key3"} == {"key2", "key3"} ^ d.keys()

    def test_xor_not_implemented(self, cls: type[MultiDict[str]]) -> None:
        d = cls([("key", "value1"), ("key2", "value2")])

        sentinel_operation_result = object()

        class RightOperand:
            def __rxor__(self, other: KeysView[str]) -> object:
                assert isinstance(other, KeysView)
                return sentinel_operation_result

        assert d.keys() ^ RightOperand() is sentinel_operation_result

    def test_xor_iterable_not_set(self, cls: type[MultiDict[str]]) -> None:
        d = cls([("key", "value1"), ("key2", "value2")])

        assert {"key", "key3"} == d.keys() ^ ["key2", "key3"]

    @pytest.mark.parametrize(
        ("key", "value", "expected"),
        (("key2", "v", True), ("key", "value1", False)),
    )
    def test_isdisjoint(
        self, cls: type[MultiDict[str]], key: str, value: str, expected: bool
    ) -> None:
        d = cls([("key", "value1")])
        assert d.items().isdisjoint({(key, value)}) is expected
        assert d.keys().isdisjoint({key}) is expected

    def test_repr_aiohttp_issue_410(self, cls: type[MutableMultiMapping[str]]) -> None:
        d = cls()

        try:
            raise Exception
            pytest.fail("Should never happen")  # pragma: no cover
        except Exception as e:
            repr(d)

            assert sys.exc_info()[1] == e  # noqa: PT017

    @pytest.mark.parametrize(
        "op",
        (operator.or_, operator.and_, operator.sub, operator.xor),
    )
    @pytest.mark.parametrize("other", ({"other"},))
    def test_op_issue_aiohttp_issue_410(
        self,
        cls: type[MultiDict[str]],
        op: Callable[[object, object], object],
        other: set[str],
    ) -> None:
        d = cls([("key", "value")])

        try:
            raise Exception
            pytest.fail("Should never happen")  # pragma: no cover
        except Exception as e:
            op(d.keys(), other)

            assert sys.exc_info()[1] == e  # noqa: PT017

    def test_weakref(self, cls: type[MutableMultiMapping[str]]) -> None:
        called = False

        def cb(wr: object) -> None:
            nonlocal called
            called = True

        d = cls()
        wr = weakref.ref(d, cb)
        del d
        gc.collect()
        assert called
        del wr

    def test_iter_length_hint_keys(
        self,
        cls: Union[type[MultiDict[int]], type[CIMultiDict[int]]],
    ) -> None:
        md = cls(a=1, b=2)
        it = iter(md.keys())
        assert it.__length_hint__() == 2  # type: ignore[attr-defined]

    def test_iter_length_hint_items(
        self,
        cls: Union[type[MultiDict[int]], type[CIMultiDict[int]]],
    ) -> None:
        md = cls(a=1, b=2)
        it = iter(md.items())
        assert it.__length_hint__() == 2  # type: ignore[attr-defined]

    def test_iter_length_hint_values(
        self,
        cls: Union[type[MultiDict[int]], type[CIMultiDict[int]]],
    ) -> None:
        md = cls(a=1, b=2)
        it = iter(md.values())
        assert it.__length_hint__() == 2

    def test_ctor_list_arg_and_kwds(
        self,
        cls: Union[type[MultiDict[int]], type[CIMultiDict[int]]],
    ) -> None:
        arg = [("a", 1)]
        obj = cls(arg, b=2)
        assert list(obj.items()) == [("a", 1), ("b", 2)]
        assert arg == [("a", 1)]

    def test_ctor_tuple_arg_and_kwds(
        self,
        cls: Union[type[MultiDict[int]], type[CIMultiDict[int]]],
    ) -> None:
        arg = (("a", 1),)
        obj = cls(arg, b=2)
        assert list(obj.items()) == [("a", 1), ("b", 2)]
        assert arg == (("a", 1),)

    def test_ctor_deque_arg_and_kwds(
        self,
        cls: Union[type[MultiDict[int]], type[CIMultiDict[int]]],
    ) -> None:
        arg = deque([("a", 1)])
        obj = cls(arg, b=2)
        assert list(obj.items()) == [("a", 1), ("b", 2)]
        assert arg == deque([("a", 1)])


class TestMultiDict(BaseMultiDictTest):
    @pytest.fixture(
        params=[
            ("MultiDict",),
            ("MultiDict", "MultiDictProxy"),
        ],
    )
    def cls(  # type: ignore[misc]
        self,
        request: pytest.FixtureRequest,
        multidict_module: ModuleType,
    ) -> Callable[..., MultiMapping[int | str] | MutableMultiMapping[int | str]]:
        """Make a case-sensitive multidict class/proxy constructor."""
        return chained_callable(multidict_module, request.param)

    def test__repr__(self, cls: type[MultiDict[str]]) -> None:
        d = cls()
        _cls = type(d)

        assert str(d) == "<%s()>" % _cls.__name__

        d = cls([("key", "one"), ("key", "two")])

        assert str(d) == "<%s('key': 'one', 'key': 'two')>" % _cls.__name__

    def test__repr___recursive(
        self, any_multidict_class: type[MultiDict[object]]
    ) -> None:
        d = any_multidict_class()
        _cls = type(d)

        d = any_multidict_class()
        d["key"] = d

        assert str(d) == "<%s('key': ...)>" % _cls.__name__

    def test_getall(self, cls: type[MultiDict[str]]) -> None:
        d = cls([("key", "value1")], key="value2")

        assert d != {"key": "value1"}
        assert len(d) == 2

        assert d.getall("key") == ["value1", "value2"]

        with pytest.raises(KeyError, match="some_key"):
            d.getall("some_key")

        default = object()
        assert d.getall("some_key", default) is default

    def test_preserve_stable_ordering(
        self,
        cls: type[MultiDict[Union[str, int]]],
    ) -> None:
        d = cls([("a", 1), ("b", "2"), ("a", 3)])
        s = "&".join("{}={}".format(k, v) for k, v in d.items())

        assert s == "a=1&b=2&a=3"

    def test_get(self, cls: type[MultiDict[int]]) -> None:
        d = cls([("a", 1), ("a", 2)])
        assert d["a"] == 1

    def test_items__repr__(self, cls: type[MultiDict[str]]) -> None:
        d = cls([("key", "value1")], key="value2")
        expected = "<_ItemsView('key': 'value1', 'key': 'value2')>"
        assert repr(d.items()) == expected

    def test_items__repr__recursive(
        self, any_multidict_class: type[MultiDict[object]]
    ) -> None:
        d = any_multidict_class()
        d["key"] = d.items()
        expected = "<_ItemsView('key': <_ItemsView('key': ...)>)>"
        assert repr(d.items()) == expected

    def test_keys__repr__(self, cls: type[MultiDict[str]]) -> None:
        d = cls([("key", "value1")], key="value2")
        assert repr(d.keys()) == "<_KeysView('key', 'key')>"

    def test_values__repr__(self, cls: type[MultiDict[str]]) -> None:
        d = cls([("key", "value1")], key="value2")
        assert repr(d.values()) == "<_ValuesView('value1', 'value2')>"

    def test_values__repr__recursive(
        self, any_multidict_class: type[MultiDict[object]]
    ) -> None:
        d = any_multidict_class()
        d["key"] = d.values()
        assert repr(d.values()) == "<_ValuesView(<_ValuesView(...)>)>"


class TestCIMultiDict(BaseMultiDictTest):
    @pytest.fixture(
        params=[
            ("CIMultiDict",),
            ("CIMultiDict", "CIMultiDictProxy"),
        ],
    )
    def cls(  # type: ignore[misc]
        self,
        request: pytest.FixtureRequest,
        multidict_module: ModuleType,
    ) -> Callable[..., MultiMapping[int | str] | MutableMultiMapping[int | str]]:
        """Make a case-insensitive multidict class/proxy constructor."""
        return chained_callable(multidict_module, request.param)

    def test_basics(self, cls: type[CIMultiDict[str]]) -> None:
        d = cls([("KEY", "value1")], KEY="value2")

        assert d.getone("key") == "value1"
        assert d.get("key") == "value1"
        assert d.get("key2", "val") == "val"
        assert d["key"] == "value1"
        assert "key" in d

        with pytest.raises(KeyError, match="key2"):
            d["key2"]
        with pytest.raises(KeyError, match="key2"):
            d.getone("key2")

    def test_from_md_and_kwds(self, cls: type[CIMultiDict[str]]) -> None:
        d = cls([("KEY", "value1")])
        d2 = cls(d, KEY="value2")

        assert list(d2.items()) == [("KEY", "value1"), ("KEY", "value2")]

    def test_getall(self, cls: type[CIMultiDict[str]]) -> None:
        d = cls([("KEY", "value1")], KEY="value2")

        assert not d == {"KEY": "value1"}
        assert len(d) == 2

        assert d.getall("key") == ["value1", "value2"]

        with pytest.raises(KeyError, match="some_key"):
            d.getall("some_key")

    def test_get(self, cls: type[CIMultiDict[int]]) -> None:
        d = cls([("A", 1), ("a", 2)])
        assert 1 == d["a"]

    def test__repr__(self, cls: type[CIMultiDict[str]]) -> None:
        d = cls([("KEY", "value1")], key="value2")
        _cls = type(d)

        expected = "<%s('KEY': 'value1', 'key': 'value2')>" % _cls.__name__
        assert str(d) == expected

    def test_items__repr__(self, cls: type[CIMultiDict[str]]) -> None:
        d = cls([("KEY", "value1")], key="value2")
        expected = "<_ItemsView('KEY': 'value1', 'key': 'value2')>"
        assert repr(d.items()) == expected

    def test_keys__repr__(self, cls: type[CIMultiDict[str]]) -> None:
        d = cls([("KEY", "value1")], key="value2")
        assert repr(d.keys()) == "<_KeysView('KEY', 'key')>"

    def test_values__repr__(self, cls: type[CIMultiDict[str]]) -> None:
        d = cls([("KEY", "value1")], key="value2")
        assert repr(d.values()) == "<_ValuesView('value1', 'value2')>"

    def test_items_iter_of_iter(self, cls: type[CIMultiDict[str]]) -> None:
        d = cls([("KEY", "value1")], key="value2")
        it = iter(d.items())
        assert iter(it) is it

    def test_keys_iter_of_iter(self, cls: type[CIMultiDict[str]]) -> None:
        d = cls([("KEY", "value1")], key="value2")
        it = iter(d.keys())
        assert iter(it) is it

    def test_values_iter_of_iter(self, cls: type[CIMultiDict[str]]) -> None:
        d = cls([("KEY", "value1")], key="value2")
        it = iter(d.values())
        assert iter(it) is it

    @pytest.mark.parametrize(
        ("arg", "expected"),
        (
            pytest.param({"key"}, {"KEY"}, id="ok"),
            pytest.param({"key", 123}, {"KEY"}, id="non-str"),
        ),
    )
    def test_keys_case_insensitive_and(
        self, cls: type[CIMultiDict[str]], arg: set[_T], expected: set[_T]
    ) -> None:
        d = cls([("KEY", "one")])
        assert d.keys() & arg == expected

    @pytest.mark.parametrize(
        ("arg", "expected"),
        (
            pytest.param(["key"], {"key"}, id="ok"),
            pytest.param(["key", 123], {"key"}, id="non-str"),
        ),
    )
    def test_keys_case_insensitive_rand(
        self, cls: type[CIMultiDict[str]], arg: list[_T], expected: set[_T]
    ) -> None:
        d = cls([("KEY", "one")])
        assert type(arg) is list
        assert arg & d.keys() == expected

    @pytest.mark.parametrize(
        ("arg", "expected"),
        (
            pytest.param({"key", "other"}, {"KEY", "other"}, id="ok"),
            pytest.param({"key", "other", 123}, {"KEY", "other", 123}, id="non-str"),
        ),
    )
    def test_keys_case_insensitive_or(
        self, cls: type[CIMultiDict[str]], arg: set[_T], expected: set[_T]
    ) -> None:
        d = cls([("KEY", "one")])

        assert d.keys() | arg == expected

    @pytest.mark.parametrize(
        ("arg", "expected"),
        (
            pytest.param(["key", "other"], {"key", "other"}, id="ok"),
            pytest.param(["key", "other", 123], {"key", "other", 123}, id="non-str"),
        ),
    )
    def test_keys_case_insensitive_ror(
        self, cls: type[CIMultiDict[str]], arg: list[_T], expected: set[_T]
    ) -> None:
        d = cls([("KEY", "one")])
        assert type(arg) is list

        assert arg | d.keys() == expected

    @pytest.mark.parametrize(
        ("arg", "expected"),
        (
            pytest.param({"key", "other"}, {"KEY2"}, id="ok"),
            pytest.param({"key", "other", 123}, {"KEY2"}, id="non-str"),
        ),
    )
    def test_keys_case_insensitive_sub(
        self, cls: type[CIMultiDict[str]], arg: set[_T], expected: set[_T]
    ) -> None:
        d = cls([("KEY", "one"), ("KEY2", "two")])

        assert d.keys() - arg == expected

    @pytest.mark.parametrize(
        ("arg", "expected"),
        (
            pytest.param(["key", "other"], {"other"}, id="ok"),
            pytest.param(["key", "other", 123], {"other", 123}, id="non-str"),
        ),
    )
    def test_keys_case_insensitive_rsub(
        self, cls: type[CIMultiDict[str]], arg: list[_T], expected: set[_T]
    ) -> None:
        d = cls([("KEY", "one"), ("KEY2", "two")])
        assert type(arg) is list

        assert arg - d.keys() == expected

    @pytest.mark.parametrize(
        ("arg", "expected"),
        (
            pytest.param(["key", "other"], {"KEY2", "other"}, id="ok"),
            pytest.param(["key", "other", 123], {"KEY2", "other", 123}, id="non-str"),
        ),
    )
    def test_keys_case_insensitive_xor(
        self, cls: type[CIMultiDict[str]], arg: list[_T], expected: set[_T]
    ) -> None:
        d = cls([("KEY", "one"), ("KEY2", "two")])

        assert d.keys() ^ arg == expected

    @pytest.mark.parametrize(
        ("arg", "expected"),
        (
            pytest.param(["key", "other"], {"KEY2", "other"}, id="ok"),
            pytest.param(["key", "other", 123], {"KEY2", "other", 123}, id="non-str"),
        ),
    )
    def test_keys_case_insensitive_rxor(
        self, cls: type[CIMultiDict[str]], arg: list[_T], expected: set[_T]
    ) -> None:
        d = cls([("KEY", "one"), ("KEY2", "two")])

        assert arg ^ d.keys() == expected

    @pytest.mark.parametrize(
        ("arg", "expected"),
        (
            pytest.param({"key"}, False, id="ok"),
            pytest.param({123}, True, id="non-str"),
        ),
    )
    def test_keys_case_insensitive_isdisjoint(
        self, cls: type[CIMultiDict[str]], arg: set[_T], expected: bool
    ) -> None:
        d = cls([("KEY", "one")])
        assert d.keys().isdisjoint(arg) == expected

    def test_keys_case_insensitive_not_iterable(
        self, cls: type[CIMultiDict[str]]
    ) -> None:
        d = cls([("KEY", "one"), ("KEY2", "two")])

        with pytest.raises(TypeError):
            123 & d.keys()  # type: ignore[operator]

        with pytest.raises(TypeError):
            d.keys() & 123  # type: ignore[operator]

        with pytest.raises(TypeError):
            123 | d.keys()  # type: ignore[operator]

        with pytest.raises(TypeError):
            d.keys() | 123  # type: ignore[operator]

        with pytest.raises(TypeError):
            123 ^ d.keys()  # type: ignore[operator]

        with pytest.raises(TypeError):
            d.keys() ^ 123  # type: ignore[operator]

        with pytest.raises(TypeError):
            d.keys() - 123  # type: ignore[operator]

        with pytest.raises(TypeError):
            123 - d.keys()  # type: ignore[operator]

    @pytest.mark.parametrize(
        "param",
        (
            pytest.param("non-tuple", id="not-tuple"),
            pytest.param(("key2", "two", "three"), id="not-2-elems"),
            pytest.param((123, "two"), id="not-str"),
        ),
    )
    def test_items_case_insensitive_parse_item(
        self, cls: type[CIMultiDict[str]], param: _T
    ) -> None:
        d = cls([("KEY", "one")])
        assert d.items() | {param} == {("KEY", "one"), param}

    @pytest.mark.parametrize(
        ("arg", "expected"),
        (
            pytest.param({("key", "one")}, {("KEY", "one")}, id="ok"),
            pytest.param(
                {("key", "one"), (123, "two")},
                {("KEY", "one")},
                id="non-str",
            ),
            pytest.param(
                {("key", "one"), ("key", "two")},
                {("KEY", "one")},
                id="nonequal-value",
            ),
        ),
    )
    def test_items_case_insensitive_and(
        self, cls: type[CIMultiDict[str]], arg: set[_T], expected: set[_T]
    ) -> None:
        d = cls([("KEY", "one")])
        assert d.items() & arg == expected

    @pytest.mark.parametrize(
        ("arg", "expected"),
        (
            pytest.param([("key", "one")], {("key", "one")}, id="ok"),
            pytest.param(
                [("key", "one"), (123, "two")],
                {("key", "one")},
                id="non-str",
            ),
            pytest.param(
                [("key", "one"), ("key", "two")],
                {("key", "one")},
                id="nonequal-value",
            ),
        ),
    )
    def test_items_case_insensitive_rand(
        self, cls: type[CIMultiDict[str]], arg: list[_T], expected: set[_T]
    ) -> None:
        d = cls([("KEY", "one")])
        assert type(arg) is list
        assert arg & d.items() == expected

    def test_items_case_insensitive_or(self, cls: type[CIMultiDict[str]]) -> None:
        d = cls([("KEY", "one")])

        assert d.items() | {("key", "one"), ("other", "two")} == {
            ("KEY", "one"),
            ("other", "two"),
        }

    def test_items_case_insensitive_ror(self, cls: type[CIMultiDict[str]]) -> None:
        d = cls([("KEY", "one"), ("KEY2", "three")])

        assert [("key", "one"), ("other", "two")] | d.items() == {
            ("key", "one"),
            ("other", "two"),
            ("KEY2", "three"),
        }

    @pytest.mark.parametrize(
        ("arg", "expected"),
        (
            pytest.param(
                {("key", "one"), ("other", "three")}, {("KEY2", "two")}, id="ok"
            ),
            pytest.param(
                {("key", "one"), (123, "three")}, {("KEY2", "two")}, id="non-str"
            ),
        ),
    )
    def test_items_case_insensitive_sub(
        self, cls: type[CIMultiDict[str]], arg: set[_T], expected: set[_T]
    ) -> None:
        d = cls([("KEY", "one"), ("KEY2", "two")])

        assert d.items() - arg == expected

    @pytest.mark.parametrize(
        ("arg", "expected"),
        (
            pytest.param(
                [("key", "one"), ("other", "three")], {("other", "three")}, id="ok"
            ),
            pytest.param(
                [("key", "one"), (123, "three")], {(123, "three")}, id="non-str"
            ),
        ),
    )
    def test_items_case_insensitive_rsub(
        self, cls: type[CIMultiDict[str]], arg: set[_T], expected: set[_T]
    ) -> None:
        d = cls([("KEY", "one"), ("KEY2", "two")])

        assert arg - d.items() == expected

    @pytest.mark.parametrize(
        ("arg", "expected"),
        (
            pytest.param(
                {("key", "one"), ("other", "three")},
                {("KEY2", "two"), ("other", "three")},
                id="ok",
            ),
            pytest.param(
                {("key", "one"), (123, "three")},
                {("KEY2", "two"), (123, "three")},
                id="non-str",
            ),
        ),
    )
    def test_items_case_insensitive_xor(
        self, cls: type[CIMultiDict[str]], arg: set[_T], expected: set[_T]
    ) -> None:
        d = cls([("KEY", "one"), ("KEY2", "two")])

        assert d.items() ^ arg == expected

    def test_items_case_insensitive_rxor(self, cls: type[CIMultiDict[str]]) -> None:
        d = cls([("KEY", "one"), ("KEY2", "two")])

        assert [("key", "one"), ("other", "three")] ^ d.items() == {
            ("KEY2", "two"),
            ("other", "three"),
        }

    def test_items_case_insensitive_non_iterable(
        self, cls: type[CIMultiDict[str]]
    ) -> None:
        d = cls([("KEY", "one")])

        with pytest.raises(TypeError):
            d.items() & None  # type: ignore[operator]

        with pytest.raises(TypeError):
            None & d.items()  # type: ignore[operator]

        with pytest.raises(TypeError):
            d.items() | None  # type: ignore[operator]

        with pytest.raises(TypeError):
            None | d.items()  # type: ignore[operator]

        with pytest.raises(TypeError):
            d.items() ^ None  # type: ignore[operator]

        with pytest.raises(TypeError):
            None ^ d.items()  # type: ignore[operator]

        with pytest.raises(TypeError):
            d.items() - None  # type: ignore[operator]

        with pytest.raises(TypeError):
            None - d.items()  # type: ignore[operator]

    @pytest.mark.parametrize(
        ("arg", "expected"),
        (
            pytest.param({("key", "one")}, False, id="ok"),
            pytest.param({(123, "one")}, True, id="non-str"),
        ),
    )
    def test_items_case_insensitive_isdisjoint(
        self, cls: type[CIMultiDict[str]], arg: set[_T], expected: bool
    ) -> None:
        d = cls([("KEY", "one")])
        assert d.items().isdisjoint(arg) == expected


def test_create_multidict_from_existing_multidict_new_pairs() -> None:
    """Test creating a MultiDict from an existing one does not mutate the original."""
    original = MultiDict([("h1", "header1"), ("h2", "header2"), ("h3", "header3")])
    new = MultiDict(original, h4="header4")
    assert "h4" in new
    assert "h4" not in original


def test_convert_multidict_to_cimultidict_and_back(
    case_sensitive_multidict_class: type[MultiDict[str]],
    case_insensitive_multidict_class: type[CIMultiDict[str]],
    case_insensitive_str_class: type[istr],
) -> None:
    """Test conversion from MultiDict to CIMultiDict."""
    start_as_md = case_sensitive_multidict_class(
        [("KEY", "value1"), ("key2", "value2")]
    )
    assert start_as_md.get("KEY") == "value1"
    assert start_as_md["KEY"] == "value1"
    assert start_as_md.get("key2") == "value2"
    assert start_as_md["key2"] == "value2"
    start_as_cimd = case_insensitive_multidict_class(
        [("KEY", "value1"), ("key2", "value2")]
    )
    assert start_as_cimd.get("key") == "value1"
    assert start_as_cimd["key"] == "value1"
    assert start_as_cimd.get("key2") == "value2"
    assert start_as_cimd["key2"] == "value2"
    converted_to_ci = case_insensitive_multidict_class(start_as_md)
    assert converted_to_ci.get("key") == "value1"
    assert converted_to_ci["key"] == "value1"
    assert converted_to_ci.get("key2") == "value2"
    assert converted_to_ci["key2"] == "value2"
    converted_to_md = case_sensitive_multidict_class(converted_to_ci)
    assert all(type(k) is case_insensitive_str_class for k in converted_to_ci.keys())
    assert converted_to_md.get("KEY") == "value1"
    assert converted_to_md["KEY"] == "value1"
    assert converted_to_md.get("key2") == "value2"
    assert converted_to_md["key2"] == "value2"


def test_convert_multidict_to_cimultidict_eq(
    case_sensitive_multidict_class: type[MultiDict[str]],
    case_insensitive_multidict_class: type[CIMultiDict[str]],
) -> None:
    """Test compare after conversion from MultiDict to CIMultiDict."""
    original = case_sensitive_multidict_class(
        [("h1", "header1"), ("h2", "header2"), ("h3", "header3")]
    )
    assert case_insensitive_multidict_class(
        original
    ) == case_insensitive_multidict_class(
        [("H1", "header1"), ("H2", "header2"), ("H3", "header3")]
    )


@pytest.mark.skipif(IS_PYPY, reason="getrefcount is not supported on PyPy")
def test_extend_does_not_alter_refcount(
    case_sensitive_multidict_class: type[MultiDict[str]],
) -> None:
    """Test that extending a MultiDict with a MultiDict does not alter the refcount of the original."""
    original = case_sensitive_multidict_class([("h1", "header1")])
    new = case_sensitive_multidict_class([("h2", "header2")])
    original_refcount = sys.getrefcount(original)
    new.extend(original)
    assert sys.getrefcount(original) == original_refcount


@pytest.mark.skipif(IS_PYPY, reason="getrefcount is not supported on PyPy")
def test_update_does_not_alter_refcount(
    case_sensitive_multidict_class: type[MultiDict[str]],
) -> None:
    """Test that updating a MultiDict with a MultiDict does not alter the refcount of the original."""
    original = case_sensitive_multidict_class([("h1", "header1")])
    new = case_sensitive_multidict_class([("h2", "header2")])
    original_refcount = sys.getrefcount(original)
    new.update(original)
    assert sys.getrefcount(original) == original_refcount


@pytest.mark.skipif(IS_PYPY, reason="getrefcount is not supported on PyPy")
def test_init_does_not_alter_refcount(
    case_sensitive_multidict_class: type[MultiDict[str]],
) -> None:
    """Test that initializing a MultiDict with a MultiDict does not alter the refcount of the original."""
    original = case_sensitive_multidict_class([("h1", "header1")])
    original_refcount = sys.getrefcount(original)
    case_sensitive_multidict_class(original)
    assert sys.getrefcount(original) == original_refcount


def test_subclassed_multidict(
    any_multidict_class: type[MultiDict[str]],
) -> None:
    """Test that subclassed MultiDicts work as expected."""
    class SubclassedMultiDict(any_multidict_class):  # type: ignore[valid-type, misc]
        """Subclassed MultiDict."""

    d1 = SubclassedMultiDict([("key", "value1")])
    d2 = SubclassedMultiDict([("key", "value2")])
    d3 = SubclassedMultiDict([("key", "value1")])
    assert d1 != d2
    assert d1 == d3
    assert d1 == SubclassedMultiDict([("key", "value1")])
    assert d1 != SubclassedMultiDict([("key", "value2")])



import pathlib
import platform
import subprocess
import sys

import pytest

IS_PYPY = platform.python_implementation() == "PyPy"


@pytest.mark.parametrize(
    ("script"),
    (
        "multidict_extend_dict.py",
        "multidict_extend_multidict.py",
        "multidict_extend_tuple.py",
        "multidict_update_multidict.py",
    ),
)
@pytest.mark.leaks
@pytest.mark.skipif(IS_PYPY, reason="leak testing is not supported on PyPy")
def test_leak(script: str) -> None:
    """Run isolated leak test script and check for leaks."""
    leak_test_script = pathlib.Path(__file__).parent.joinpath("isolated", script)

    subprocess.run(
        [sys.executable, "-u", str(leak_test_script)],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        check=True,
    )



"""Tests for circular imports in all local packages and modules.

This ensures all internal packages can be imported right away without
any need to import some other module before doing so.

This module is based on the idea that pytest uses for self-testing:
* https://github.com/sanitizers/octomachinery/blob/be18b54/tests/circular_imports_test.py  # noqa: E501
* https://github.com/pytest-dev/pytest/blob/d18c75b/testing/test_meta.py
* https://twitter.com/codewithanthony/status/1229445110510735361
"""

from __future__ import annotations

import os
import pkgutil
import subprocess
import sys
from itertools import chain
from pathlib import Path
from types import ModuleType
from typing import Generator

import pytest

import multidict


def _find_all_importables(pkg: ModuleType) -> list[str]:
    """Find all importables in the project.

    Return them in order.
    """
    return sorted(
        set(
            chain.from_iterable(
                _discover_path_importables(Path(p), pkg.__name__) for p in pkg.__path__
            ),
        ),
    )


def _discover_path_importables(
    pkg_pth: Path,
    pkg_name: str,
) -> Generator[str, None, None]:
    """Yield all importables under a given path and package."""
    yield pkg_name
    for dir_path, _d, file_names in os.walk(pkg_pth):
        pkg_dir_path = Path(dir_path)

        if pkg_dir_path.parts[-1] == "__pycache__":
            continue

        rel_pt = pkg_dir_path.relative_to(pkg_pth)
        pkg_pref = ".".join((pkg_name,) + rel_pt.parts)
        yield from (
            pkg_path
            for _, pkg_path, _ in pkgutil.walk_packages(
                (str(pkg_dir_path),),
                prefix=f"{pkg_pref}.",
            )
        )


@pytest.fixture(params=_find_all_importables(multidict))
def import_path(request: pytest.FixtureRequest) -> str:
    """Return an importable from the multidict package."""
    importable_module: str = request.param
    if importable_module == "multidict._multidict":
        request.applymarker(pytest.mark.c_extension)

    return importable_module


def test_no_warnings(import_path: str) -> None:
    """Verify that importing modules and packages doesn't explode.

    This is seeking for any import errors including ones caused
    by circular imports.
    """
    imp_cmd = (
        # fmt: off
        sys.executable,
        "-I",
        "-W", "error",
        "-c", f"import {import_path!s}",
        # fmt: on
    )

    subprocess.check_call(imp_cmd)


@pytest.mark.c_extension
def test_c_extension_preferred_by_default(monkeypatch: pytest.MonkeyPatch) -> None:
    """Verify that the C-extension is exposed by default."""
    monkeypatch.delenv("MULTIDICT_NO_EXTENSIONS", raising=False)

    imp_cmd = (
        # fmt: off
        sys.executable,
        "-I",
        "-W", "error",
        "-c", "import multidict; raise SystemExit(int("
        "multidict.istr.__module__ != 'multidict._multidict' "
        "or multidict.USE_EXTENSIONS is not True))",
        # fmt: on
    )

    subprocess.check_call(imp_cmd)



import pytest

from multidict import MultiDict


def test_guard_items(
    case_sensitive_multidict_class: type[MultiDict[str]],
) -> None:
    md = case_sensitive_multidict_class({"a": "b"})
    it = iter(md.items())
    md["a"] = "c"
    with pytest.raises(RuntimeError):
        next(it)


def test_guard_keys(
    case_sensitive_multidict_class: type[MultiDict[str]],
) -> None:
    md = case_sensitive_multidict_class({"a": "b"})
    it = iter(md.keys())
    md["a"] = "c"
    with pytest.raises(RuntimeError):
        next(it)


def test_guard_values(
    case_sensitive_multidict_class: type[MultiDict[str]],
) -> None:
    md = case_sensitive_multidict_class({"a": "b"})
    it = iter(md.values())
    md["a"] = "c"
    with pytest.raises(RuntimeError):
        next(it)



"""codspeed benchmarks for multidict."""

from typing import Dict, Type, Union

from pytest_codspeed import BenchmarkFixture

from multidict import (
    CIMultiDict,
    CIMultiDictProxy,
    MultiDict,
    MultiDictProxy,
    istr,
)

# Note that this benchmark should not be refactored to use pytest.mark.parametrize
# since each benchmark name should be unique.

_SENTINEL = object()


def test_multidict_insert_str(
    benchmark: BenchmarkFixture, any_multidict_class: Type[MultiDict[str]]
) -> None:
    md = any_multidict_class()
    items = [str(i) for i in range(100)]

    @benchmark
    def _run() -> None:
        for i in items:
            md[i] = i


def test_cimultidict_insert_istr(
    benchmark: BenchmarkFixture,
    case_insensitive_multidict_class: Type[CIMultiDict[istr]],
) -> None:
    md = case_insensitive_multidict_class()
    items = [istr(i) for i in range(100)]

    @benchmark
    def _run() -> None:
        for i in items:
            md[i] = i


def test_multidict_add_str(
    benchmark: BenchmarkFixture, any_multidict_class: Type[MultiDict[str]]
) -> None:
    base_md = any_multidict_class()
    items = [str(i) for i in range(100)]

    @benchmark
    def _run() -> None:
        for _ in range(100):
            md = base_md.copy()
            for i in items:
                md.add(i, i)


def test_cimultidict_add_istr(
    benchmark: BenchmarkFixture,
    case_insensitive_multidict_class: Type[CIMultiDict[istr]],
) -> None:
    base_md = case_insensitive_multidict_class()
    items = [istr(i) for i in range(100)]

    @benchmark
    def _run() -> None:
        for j in range(100):
            md = base_md.copy()
            for i in items:
                md.add(i, i)


def test_multidict_pop_str(
    benchmark: BenchmarkFixture, any_multidict_class: Type[MultiDict[str]]
) -> None:
    md_base = any_multidict_class((str(i), str(i)) for i in range(100))
    items = [str(i) for i in range(100)]

    @benchmark
    def _run() -> None:
        md = md_base.copy()
        for i in items:
            md.pop(i)


def test_cimultidict_pop_istr(
    benchmark: BenchmarkFixture,
    case_insensitive_multidict_class: Type[CIMultiDict[istr]],
) -> None:
    md_base = case_insensitive_multidict_class((istr(i), istr(i)) for i in range(100))
    items = [istr(i) for i in range(100)]

    @benchmark
    def _run() -> None:
        md = md_base.copy()
        for i in items:
            md.pop(i)


def test_multidict_popitem_str(
    benchmark: BenchmarkFixture, any_multidict_class: Type[MultiDict[str]]
) -> None:
    md_base = any_multidict_class((str(i), str(i)) for i in range(100))

    @benchmark
    def _run() -> None:
        md = md_base.copy()
        for _ in range(100):
            md.popitem()


def test_multidict_clear_str(
    benchmark: BenchmarkFixture, any_multidict_class: Type[MultiDict[str]]
) -> None:
    md = any_multidict_class((str(i), str(i)) for i in range(100))

    @benchmark
    def _run() -> None:
        md.clear()


def test_multidict_update_str(
    benchmark: BenchmarkFixture, any_multidict_class: Type[MultiDict[str]]
) -> None:
    md = any_multidict_class((str(i), str(i)) for i in range(100))
    items = {str(i): str(i) for i in range(100, 200)}

    @benchmark
    def _run() -> None:
        md.update(items)


def test_cimultidict_update_istr(
    benchmark: BenchmarkFixture,
    case_insensitive_multidict_class: Type[CIMultiDict[istr]],
) -> None:
    md = case_insensitive_multidict_class((istr(i), istr(i)) for i in range(100))
    items: Dict[Union[str, istr], istr] = {istr(i): istr(i) for i in range(100, 200)}

    @benchmark
    def _run() -> None:
        md.update(items)


def test_multidict_update_str_with_kwargs(
    benchmark: BenchmarkFixture, any_multidict_class: Type[MultiDict[str]]
) -> None:
    md = any_multidict_class((str(i), str(i)) for i in range(100))
    items = {str(i): str(i) for i in range(100, 200)}
    kwargs = {str(i): str(i) for i in range(200, 300)}

    @benchmark
    def _run() -> None:
        md.update(items, **kwargs)


def test_cimultidict_update_istr_with_kwargs(
    benchmark: BenchmarkFixture,
    case_insensitive_multidict_class: Type[CIMultiDict[istr]],
) -> None:
    md = case_insensitive_multidict_class((istr(i), istr(i)) for i in range(100))
    items: Dict[Union[str, istr], istr] = {istr(i): istr(i) for i in range(100, 200)}
    kwargs = {str(i): istr(i) for i in range(200, 300)}

    @benchmark
    def _run() -> None:
        md.update(items, **kwargs)


def test_multidict_extend_str(
    benchmark: BenchmarkFixture, any_multidict_class: Type[MultiDict[str]]
) -> None:
    base_md = any_multidict_class((str(i), str(i)) for i in range(100))
    items = {str(i): str(i) for i in range(200)}

    @benchmark
    def _run() -> None:
        for j in range(100):
            md = base_md.copy()
            md.extend(items)


def test_cimultidict_extend_istr(
    benchmark: BenchmarkFixture,
    case_insensitive_multidict_class: Type[CIMultiDict[istr]],
) -> None:
    base_md = case_insensitive_multidict_class((istr(i), istr(i)) for i in range(100))
    items = {istr(i): istr(i) for i in range(200)}

    @benchmark
    def _run() -> None:
        for _ in range(100):
            md = base_md.copy()
            md.extend(items)


def test_multidict_extend_str_with_kwargs(
    benchmark: BenchmarkFixture, any_multidict_class: Type[MultiDict[str]]
) -> None:
    base_md = any_multidict_class((str(i), str(i)) for i in range(100))
    items = {str(i): str(i) for i in range(200)}
    kwargs = {str(i): str(i) for i in range(200, 300)}

    @benchmark
    def _run() -> None:
        for j in range(100):
            md = base_md.copy()
            md.extend(items, **kwargs)


def test_cimultidict_extend_istr_with_kwargs(
    benchmark: BenchmarkFixture,
    case_insensitive_multidict_class: Type[CIMultiDict[istr]],
) -> None:
    base_md = case_insensitive_multidict_class((istr(i), istr(i)) for i in range(100))
    items = {istr(i): istr(i) for i in range(200)}
    kwargs = {str(i): istr(i) for i in range(200, 300)}

    @benchmark
    def _run() -> None:
        for _ in range(100):
            md = base_md.copy()
            md.extend(items, **kwargs)


def test_multidict_delitem_str(
    benchmark: BenchmarkFixture, any_multidict_class: Type[MultiDict[str]]
) -> None:
    md_base = any_multidict_class((str(i), str(i)) for i in range(100))
    items = [str(i) for i in range(100)]

    @benchmark
    def _run() -> None:
        md = md_base.copy()
        for i in items:
            del md[i]


def test_cimultidict_delitem_istr(
    benchmark: BenchmarkFixture,
    case_insensitive_multidict_class: Type[CIMultiDict[istr]],
) -> None:
    md_base = case_insensitive_multidict_class((istr(i), istr(i)) for i in range(100))
    items = [istr(i) for i in range(100)]

    @benchmark
    def _run() -> None:
        md = md_base.copy()
        for i in items:
            del md[i]


def test_multidict_getall_str_hit(
    benchmark: BenchmarkFixture, any_multidict_class: Type[MultiDict[str]]
) -> None:
    md = any_multidict_class((f"key{j}", str(f"{i}-{j}"))
                             for i in range(20) for j in range(5))

    @benchmark
    def _run() -> None:
        md.getall("key0")


def test_multidict_getall_str_miss(
    benchmark: BenchmarkFixture, any_multidict_class: Type[MultiDict[str]]
) -> None:
    md = any_multidict_class((f"key{j}", str(f"{i}-{j}"))
                             for i in range(20) for j in range(5))

    @benchmark
    def _run() -> None:
        md.getall("miss", ())


def test_cimultidict_getall_istr_hit(
    benchmark: BenchmarkFixture,
    case_insensitive_multidict_class: Type[CIMultiDict[istr]],
) -> None:
    all_istr = istr("key0")
    md = case_insensitive_multidict_class((f"key{j}", istr(f"{i}-{j}"))
                                          for i in range(20) for j in range(5))

    @benchmark
    def _run() -> None:
        md.getall(all_istr)


def test_cimultidict_getall_istr_miss(
    benchmark: BenchmarkFixture,
    case_insensitive_multidict_class: Type[CIMultiDict[istr]],
) -> None:
    miss_istr = istr("miss")
    md = case_insensitive_multidict_class((istr(f"key{j}"), istr(f"{i}-{j}"))
                                          for i in range(20) for j in range(5))

    @benchmark
    def _run() -> None:
        md.getall(miss_istr, ())


def test_multidict_fetch(
    benchmark: BenchmarkFixture, any_multidict_class: Type[MultiDict[str]]
) -> None:
    md = any_multidict_class((str(i), str(i)) for i in range(100))
    items = [str(i) for i in range(100)]

    @benchmark
    def _run() -> None:
        for i in items:
            md[i]


def test_cimultidict_fetch_istr(
    benchmark: BenchmarkFixture,
    case_insensitive_multidict_class: Type[CIMultiDict[istr]],
) -> None:
    md = case_insensitive_multidict_class((istr(i), istr(i)) for i in range(100))
    items = [istr(i) for i in range(100)]

    @benchmark
    def _run() -> None:
        for i in items:
            md[i]


def test_multidict_get_hit(
    benchmark: BenchmarkFixture, any_multidict_class: Type[MultiDict[str]]
) -> None:
    md = any_multidict_class((str(i), str(i)) for i in range(100))
    items = [str(i) for i in range(100)]

    @benchmark
    def _run() -> None:
        for i in items:
            md.get(i)


def test_multidict_get_miss(
    benchmark: BenchmarkFixture, any_multidict_class: Type[MultiDict[str]]
) -> None:
    md = any_multidict_class((str(i), str(i)) for i in range(100))
    items = [str(i) for i in range(100, 200)]

    @benchmark
    def _run() -> None:
        for i in items:
            md.get(i)


def test_cimultidict_get_istr_hit(
    benchmark: BenchmarkFixture,
    case_insensitive_multidict_class: Type[CIMultiDict[istr]],
) -> None:
    md = case_insensitive_multidict_class((istr(i), istr(i)) for i in range(100))
    items = [istr(i) for i in range(100)]

    @benchmark
    def _run() -> None:
        for i in items:
            md.get(i)


def test_cimultidict_get_istr_miss(
    benchmark: BenchmarkFixture,
    case_insensitive_multidict_class: Type[CIMultiDict[istr]],
) -> None:
    md = case_insensitive_multidict_class((istr(i), istr(i)) for i in range(100))
    items = [istr(i) for i in range(100, 200)]

    @benchmark
    def _run() -> None:
        for i in items:
            md.get(i)


def test_multidict_get_hit_with_default(
    benchmark: BenchmarkFixture, any_multidict_class: Type[MultiDict[str]]
) -> None:
    md = any_multidict_class((str(i), str(i)) for i in range(100))
    items = [str(i) for i in range(100)]

    @benchmark
    def _run() -> None:
        for i in items:
            md.get(i, _SENTINEL)


def test_cimultidict_get_istr_hit_with_default(
    benchmark: BenchmarkFixture,
    case_insensitive_multidict_class: Type[CIMultiDict[istr]],
) -> None:
    md = case_insensitive_multidict_class((istr(i), istr(i)) for i in range(100))
    items = [istr(i) for i in range(100)]

    @benchmark
    def _run() -> None:
        for i in items:
            md.get(i, _SENTINEL)


def test_cimultidict_get_istr_with_default_miss(
    benchmark: BenchmarkFixture,
    case_insensitive_multidict_class: Type[CIMultiDict[istr]],
) -> None:
    md = case_insensitive_multidict_class((istr(i), istr(i)) for i in range(100))
    items = [istr(i) for i in range(100, 200)]

    @benchmark
    def _run() -> None:
        for i in items:
            md.get(i, _SENTINEL)


def test_multidict_repr(
    benchmark: BenchmarkFixture, any_multidict_class: Type[MultiDict[str]]
) -> None:
    items = [str(i) for i in range(100)]
    md = any_multidict_class([(i, i) for i in items])

    @benchmark
    def _run() -> None:
        repr(md)


def test_create_empty_multidict(
    benchmark: BenchmarkFixture, any_multidict_class: Type[MultiDict[str]]
) -> None:
    @benchmark
    def _run() -> None:
        any_multidict_class()


def test_create_multidict_with_items(
    benchmark: BenchmarkFixture, any_multidict_class: Type[MultiDict[str]]
) -> None:
    items = [(str(i), str(i)) for i in range(100)]

    @benchmark
    def _run() -> None:
        any_multidict_class(items)


def test_create_cimultidict_with_items_istr(
    benchmark: BenchmarkFixture,
    case_insensitive_multidict_class: Type[CIMultiDict[istr]],
) -> None:
    items = [(istr(i), istr(i)) for i in range(100)]

    @benchmark
    def _run() -> None:
        case_insensitive_multidict_class(items)


def test_create_multidict_with_dict(
    benchmark: BenchmarkFixture, any_multidict_class: Type[MultiDict[str]]
) -> None:
    dct = {str(i): str(i) for i in range(100)}

    @benchmark
    def _run() -> None:
        any_multidict_class(dct)


def test_create_cimultidict_with_dict_istr(
    benchmark: BenchmarkFixture,
    case_insensitive_multidict_class: Type[CIMultiDict[istr]],
) -> None:
    dct = {istr(i): istr(i) for i in range(100)}

    @benchmark
    def _run() -> None:
        case_insensitive_multidict_class(dct)


def test_create_multidict_with_items_with_kwargs(
    benchmark: BenchmarkFixture, any_multidict_class: Type[MultiDict[str]]
) -> None:
    items = [(str(i), str(i)) for i in range(100)]
    kwargs = {str(i): str(i) for i in range(100)}

    @benchmark
    def _run() -> None:
        any_multidict_class(items, **kwargs)


def test_create_cimultidict_with_items_istr_with_kwargs(
    benchmark: BenchmarkFixture,
    case_insensitive_multidict_class: Type[CIMultiDict[istr]],
) -> None:
    items = [(istr(i), istr(i)) for i in range(100)]
    kwargs = {str(i): istr(i) for i in range(100)}

    @benchmark
    def _run() -> None:
        case_insensitive_multidict_class(items, **kwargs)


def test_create_empty_multidictproxy(benchmark: BenchmarkFixture) -> None:
    md: MultiDict[str] = MultiDict()

    @benchmark
    def _run() -> None:
        MultiDictProxy(md)


def test_create_multidictproxy(benchmark: BenchmarkFixture) -> None:
    items = [(str(i), str(i)) for i in range(100)]
    md: MultiDict[str] = MultiDict(items)

    @benchmark
    def _run() -> None:
        MultiDictProxy(md)


def test_create_empty_cimultidictproxy(
    benchmark: BenchmarkFixture,
) -> None:
    md: CIMultiDict[istr] = CIMultiDict()

    @benchmark
    def _run() -> None:
        CIMultiDictProxy(md)


def test_create_cimultidictproxy(
    benchmark: BenchmarkFixture,
) -> None:
    items = [(istr(i), istr(i)) for i in range(100)]
    md = CIMultiDict(items)

    @benchmark
    def _run() -> None:
        CIMultiDictProxy(md)


def test_create_from_existing_cimultidict(
    benchmark: BenchmarkFixture,
    case_insensitive_multidict_class: Type[CIMultiDict[istr]],
) -> None:
    existing = case_insensitive_multidict_class((istr(i), istr(i)) for i in range(5))

    @benchmark
    def _run() -> None:
        case_insensitive_multidict_class(existing)


def test_copy_from_existing_cimultidict(
    benchmark: BenchmarkFixture,
    case_insensitive_multidict_class: Type[CIMultiDict[istr]],
) -> None:
    existing = case_insensitive_multidict_class((istr(i), istr(i)) for i in range(5))

    @benchmark
    def _run() -> None:
        existing.copy()


def test_iterate_multidict(
    benchmark: BenchmarkFixture, any_multidict_class: Type[MultiDict[str]]
) -> None:
    items = [(str(i), str(i)) for i in range(100)]
    md = any_multidict_class(items)

    @benchmark
    def _run() -> None:
        for _ in md:
            pass


def test_iterate_multidict_keys(
    benchmark: BenchmarkFixture, any_multidict_class: Type[MultiDict[str]]
) -> None:
    items = [(str(i), str(i)) for i in range(100)]
    md = any_multidict_class(items)

    @benchmark
    def _run() -> None:
        for _ in md.keys():
            pass


def test_iterate_multidict_values(
    benchmark: BenchmarkFixture, any_multidict_class: Type[MultiDict[str]]
) -> None:
    items = [(str(i), str(i)) for i in range(100)]
    md = any_multidict_class(items)

    @benchmark
    def _run() -> None:
        for _ in md.values():
            pass


def test_iterate_multidict_items(
    benchmark: BenchmarkFixture, any_multidict_class: Type[MultiDict[str]]
) -> None:
    items = [(str(i), str(i)) for i in range(100)]
    md = any_multidict_class(items)

    @benchmark
    def _run() -> None:
        for _, _ in md.items():
            pass



import gc
import sys
from typing import Any

import objgraph  # type: ignore[import-untyped]

from multidict import MultiDict


class NoLeakDict(dict[str, Any]):
    """A subclassed dict to make it easier to test for leaks."""


def _run_isolated_case() -> None:
    md: MultiDict[str] = MultiDict()
    for _ in range(100):
        md.update(NoLeakDict())
    del md
    gc.collect()

    leaked = len(objgraph.by_type("NoLeakDict"))
    print(f"{leaked} instances of NoLeakDict not collected by GC")
    sys.exit(1 if leaked else 0)


if __name__ == "__main__":
    _run_isolated_case()



import types

import pytest


def test_proxies(multidict_module: types.ModuleType) -> None:
    assert issubclass(
        multidict_module.CIMultiDictProxy,
        multidict_module.MultiDictProxy,
    )


def test_dicts(multidict_module: types.ModuleType) -> None:
    assert issubclass(multidict_module.CIMultiDict, multidict_module.MultiDict)


def test_proxy_not_inherited_from_dict(multidict_module: types.ModuleType) -> None:
    assert not issubclass(multidict_module.MultiDictProxy, multidict_module.MultiDict)


def test_dict_not_inherited_from_proxy(multidict_module: types.ModuleType) -> None:
    assert not issubclass(multidict_module.MultiDict, multidict_module.MultiDictProxy)


def test_multidict_proxy_copy_type(multidict_module: types.ModuleType) -> None:
    d = multidict_module.MultiDict(key="val")
    p = multidict_module.MultiDictProxy(d)
    assert isinstance(p.copy(), multidict_module.MultiDict)


def test_cimultidict_proxy_copy_type(multidict_module: types.ModuleType) -> None:
    d = multidict_module.CIMultiDict(key="val")
    p = multidict_module.CIMultiDictProxy(d)
    assert isinstance(p.copy(), multidict_module.CIMultiDict)


def test_create_multidict_proxy_from_nonmultidict(
    multidict_module: types.ModuleType,
) -> None:
    with pytest.raises(TypeError):
        multidict_module.MultiDictProxy({})


def test_create_multidict_proxy_from_cimultidict(
    multidict_module: types.ModuleType,
) -> None:
    d = multidict_module.CIMultiDict(key="val")
    p = multidict_module.MultiDictProxy(d)
    assert p == d


def test_create_multidict_proxy_from_multidict_proxy_from_mdict(
    multidict_module: types.ModuleType,
) -> None:
    d = multidict_module.MultiDict(key="val")
    p = multidict_module.MultiDictProxy(d)
    assert p == d
    p2 = multidict_module.MultiDictProxy(p)
    assert p2 == p


def test_create_cimultidict_proxy_from_cimultidict_proxy_from_ci(
    multidict_module: types.ModuleType,
) -> None:
    d = multidict_module.CIMultiDict(key="val")
    p = multidict_module.CIMultiDictProxy(d)
    assert p == d
    p2 = multidict_module.CIMultiDictProxy(p)
    assert p2 == p


def test_create_cimultidict_proxy_from_nonmultidict(
    multidict_module: types.ModuleType,
) -> None:
    with pytest.raises(
        TypeError,
        match=(
            "ctor requires CIMultiDict or CIMultiDictProxy instance, "
            "not <class 'dict'>"
        ),
    ):
        multidict_module.CIMultiDictProxy({})


def test_create_ci_multidict_proxy_from_multidict(
    multidict_module: types.ModuleType,
) -> None:
    d = multidict_module.MultiDict(key="val")
    with pytest.raises(
        TypeError,
        match=(
            "ctor requires CIMultiDict or CIMultiDictProxy instance, "
            "not <class 'multidict._multidict.*.MultiDict'>"
        ),
    ):
        multidict_module.CIMultiDictProxy(d)


def test_generic_alias(multidict_module: types.ModuleType) -> None:
    assert multidict_module.MultiDict[int] == types.GenericAlias(
        multidict_module.MultiDict, (int,)
    )
    assert multidict_module.MultiDictProxy[int] == types.GenericAlias(
        multidict_module.MultiDictProxy, (int,)
    )
    assert multidict_module.CIMultiDict[int] == types.GenericAlias(
        multidict_module.CIMultiDict, (int,)
    )
    assert multidict_module.CIMultiDictProxy[int] == types.GenericAlias(
        multidict_module.CIMultiDictProxy, (int,)
    )



import pickle
from pathlib import Path
from typing import TYPE_CHECKING

import pytest

from multidict import MultiDict, MultiDictProxy, istr

if TYPE_CHECKING:
    from conftest import MultidictImplementation

here = Path(__file__).resolve().parent


def test_pickle(
    any_multidict_class: type[MultiDict[int]], pickle_protocol: int
) -> None:
    d = any_multidict_class([("a", 1), ("a", 2)])
    pbytes = pickle.dumps(d, pickle_protocol)
    obj = pickle.loads(pbytes)
    assert d == obj
    assert isinstance(obj, any_multidict_class)


def test_pickle_proxy(
    any_multidict_class: type[MultiDict[int]],
    any_multidict_proxy_class: type[MultiDictProxy[int]],
) -> None:
    d = any_multidict_class([("a", 1), ("a", 2)])
    proxy = any_multidict_proxy_class(d)
    with pytest.raises(TypeError):
        pickle.dumps(proxy)


def test_pickle_istr(
    case_insensitive_str_class: type[istr], pickle_protocol: int
) -> None:
    s = case_insensitive_str_class("str")
    pbytes = pickle.dumps(s, pickle_protocol)
    obj = pickle.loads(pbytes)
    assert s == obj
    assert isinstance(obj, case_insensitive_str_class)


def test_load_from_file(
    any_multidict_class: type[MultiDict[int]],
    multidict_implementation: "MultidictImplementation",
    pickle_protocol: int,
) -> None:
    multidict_class_name = any_multidict_class.__name__
    pickle_file_basename = "-".join(
        (
            multidict_class_name.lower(),
            multidict_implementation.tag,
        )
    )
    d = any_multidict_class([("a", 1), ("a", 2)])
    fname = f"{pickle_file_basename}.pickle.{pickle_protocol}"
    p = here / fname
    with p.open("rb") as f:
        obj = pickle.load(f)
    assert d == obj
    assert isinstance(obj, any_multidict_class)


def test_load_istr_from_file(
    case_insensitive_str_class: type[istr],
    multidict_implementation: "MultidictImplementation",
    pickle_protocol: int,
) -> None:
    istr_class_name = case_insensitive_str_class.__name__
    pickle_file_basename = "-".join(
        (
            istr_class_name.lower(),
            multidict_implementation.tag,
        )
    )
    s = case_insensitive_str_class("str")
    fname = f"{pickle_file_basename}.pickle.{pickle_protocol}"
    p = here / fname
    with p.open("rb") as f:
        obj = pickle.load(f)
    assert s == obj
    assert isinstance(obj, case_insensitive_str_class)



from collections.abc import Mapping, MutableMapping

from multidict import (
    MultiDict,
    MultiDictProxy,
    MultiMapping,
    MutableMultiMapping,
)


def test_abc_inheritance() -> None:
    assert issubclass(MultiMapping, Mapping)
    assert not issubclass(MultiMapping, MutableMapping)
    assert issubclass(MutableMultiMapping, Mapping)
    assert issubclass(MutableMultiMapping, MutableMapping)


def test_multidict_inheritance(any_multidict_class: type[MultiDict[str]]) -> None:
    assert issubclass(any_multidict_class, MultiMapping)
    assert issubclass(any_multidict_class, MutableMultiMapping)


def test_proxy_inheritance(
    any_multidict_proxy_class: type[MultiDictProxy[str]],
) -> None:
    assert issubclass(any_multidict_proxy_class, MultiMapping)
    assert not issubclass(any_multidict_proxy_class, MutableMultiMapping)


def test_generic_type_in_runtime() -> None:
    MultiMapping[str]
    MutableMultiMapping[str]



"""codspeed benchmarks for multidict views."""

from typing import Type

from pytest_codspeed import BenchmarkFixture

from multidict import MultiDict


def test_keys_view_equals(
    benchmark: BenchmarkFixture, any_multidict_class: Type[MultiDict[str]]
) -> None:
    md1: MultiDict[str] = any_multidict_class({str(i): str(i) for i in range(100)})
    md2: MultiDict[str] = any_multidict_class({str(i): str(i) for i in range(100)})

    @benchmark
    def _run() -> None:
        assert md1.keys() == md2.keys()


def test_keys_view_not_equals(
    benchmark: BenchmarkFixture, any_multidict_class: Type[MultiDict[str]]
) -> None:
    md1: MultiDict[str] = any_multidict_class({str(i): str(i) for i in range(100)})
    md2: MultiDict[str] = any_multidict_class({str(i): str(i) for i in range(20, 120)})

    @benchmark
    def _run() -> None:
        assert md1.keys() != md2.keys()


def test_keys_view_more(
    benchmark: BenchmarkFixture, any_multidict_class: Type[MultiDict[str]]
) -> None:
    md: MultiDict[str] = any_multidict_class({str(i): str(i) for i in range(100)})
    s = {str(i) for i in range(50)}

    @benchmark
    def _run() -> None:
        assert md.keys() > s


def test_keys_view_more_or_equal(
    benchmark: BenchmarkFixture, any_multidict_class: Type[MultiDict[str]]
) -> None:
    md: MultiDict[str] = any_multidict_class({str(i): str(i) for i in range(100)})
    s = {str(i) for i in range(100)}

    @benchmark
    def _run() -> None:
        assert md.keys() >= s


def test_keys_view_less(
    benchmark: BenchmarkFixture, any_multidict_class: Type[MultiDict[str]]
) -> None:
    md: MultiDict[str] = any_multidict_class({str(i): str(i) for i in range(100)})
    s = {str(i) for i in range(150)}

    @benchmark
    def _run() -> None:
        assert md.keys() < s


def test_keys_view_less_or_equal(
    benchmark: BenchmarkFixture, any_multidict_class: Type[MultiDict[str]]
) -> None:
    md: MultiDict[str] = any_multidict_class({str(i): str(i) for i in range(100)})
    s = {str(i) for i in range(100)}

    @benchmark
    def _run() -> None:
        assert md.keys() <= s


def test_keys_view_and(
    benchmark: BenchmarkFixture, any_multidict_class: Type[MultiDict[str]]
) -> None:
    md1: MultiDict[str] = any_multidict_class({str(i): str(i) for i in range(100)})
    md2: MultiDict[str] = any_multidict_class({str(i): str(i) for i in range(50, 150)})

    @benchmark
    def _run() -> None:
        assert len(md1.keys() & md2.keys()) == 50


def test_keys_view_or(
    benchmark: BenchmarkFixture, any_multidict_class: Type[MultiDict[str]]
) -> None:
    md1: MultiDict[str] = any_multidict_class({str(i): str(i) for i in range(100)})
    md2: MultiDict[str] = any_multidict_class({str(i): str(i) for i in range(50, 150)})

    @benchmark
    def _run() -> None:
        assert len(md1.keys() | md2.keys()) == 150


def test_keys_view_sub(
    benchmark: BenchmarkFixture, any_multidict_class: Type[MultiDict[str]]
) -> None:
    md1: MultiDict[str] = any_multidict_class({str(i): str(i) for i in range(100)})
    md2: MultiDict[str] = any_multidict_class({str(i): str(i) for i in range(50, 150)})

    @benchmark
    def _run() -> None:
        assert len(md1.keys() - md2.keys()) == 50


def test_keys_view_xor(
    benchmark: BenchmarkFixture, any_multidict_class: Type[MultiDict[str]]
) -> None:
    md1: MultiDict[str] = any_multidict_class({str(i): str(i) for i in range(100)})
    md2: MultiDict[str] = any_multidict_class({str(i): str(i) for i in range(50, 150)})

    @benchmark
    def _run() -> None:
        assert len(md1.keys() ^ md2.keys()) == 100


def test_keys_view_is_disjoint(
    benchmark: BenchmarkFixture, any_multidict_class: Type[MultiDict[str]]
) -> None:
    md1: MultiDict[str] = any_multidict_class({str(i): str(i) for i in range(100)})
    md2: MultiDict[str] = any_multidict_class({str(i): str(i) for i in range(100, 200)})

    @benchmark
    def _run() -> None:
        assert md1.keys().isdisjoint(md2.keys())


def test_keys_view_repr(
    benchmark: BenchmarkFixture, any_multidict_class: Type[MultiDict[str]]
) -> None:
    md: MultiDict[str] = any_multidict_class({str(i): str(i) for i in range(100)})

    @benchmark
    def _run() -> None:
        repr(md.keys())


def test_items_view_equals(
    benchmark: BenchmarkFixture, any_multidict_class: Type[MultiDict[str]]
) -> None:
    md1: MultiDict[str] = any_multidict_class({str(i): str(i) for i in range(100)})
    md2: MultiDict[str] = any_multidict_class({str(i): str(i) for i in range(100)})

    @benchmark
    def _run() -> None:
        assert md1.items() == md2.items()


def test_items_view_not_equals(
    benchmark: BenchmarkFixture, any_multidict_class: Type[MultiDict[str]]
) -> None:
    md1: MultiDict[str] = any_multidict_class({str(i): str(i) for i in range(100)})
    md2: MultiDict[str] = any_multidict_class({str(i): str(i) for i in range(20, 120)})

    @benchmark
    def _run() -> None:
        assert md1.items() != md2.items()


def test_items_view_more(
    benchmark: BenchmarkFixture, any_multidict_class: Type[MultiDict[str]]
) -> None:
    md: MultiDict[str] = any_multidict_class({str(i): str(i) for i in range(100)})
    s = {(str(i), str(i)) for i in range(50)}

    @benchmark
    def _run() -> None:
        assert md.items() > s


def test_items_view_more_or_equal(
    benchmark: BenchmarkFixture, any_multidict_class: Type[MultiDict[str]]
) -> None:
    md: MultiDict[str] = any_multidict_class({str(i): str(i) for i in range(100)})
    s = {(str(i), str(i)) for i in range(100)}

    @benchmark
    def _run() -> None:
        assert md.items() >= s


def test_items_view_less(
    benchmark: BenchmarkFixture, any_multidict_class: Type[MultiDict[str]]
) -> None:
    md: MultiDict[str] = any_multidict_class({str(i): str(i) for i in range(100)})
    s = {(str(i), str(i)) for i in range(150)}

    @benchmark
    def _run() -> None:
        assert md.items() < s


def test_items_view_less_or_equal(
    benchmark: BenchmarkFixture, any_multidict_class: Type[MultiDict[str]]
) -> None:
    md: MultiDict[str] = any_multidict_class({str(i): str(i) for i in range(100)})
    s = {(str(i), str(i)) for i in range(100)}

    @benchmark
    def _run() -> None:
        assert md.items() <= s


def test_items_view_and(
    benchmark: BenchmarkFixture, any_multidict_class: Type[MultiDict[str]]
) -> None:
    md1: MultiDict[str] = any_multidict_class({str(i): str(i) for i in range(100)})
    md2: MultiDict[str] = any_multidict_class({str(i): str(i) for i in range(50, 150)})

    @benchmark
    def _run() -> None:
        assert len(md1.items() & md2.items()) == 50


def test_items_view_or(
    benchmark: BenchmarkFixture, any_multidict_class: Type[MultiDict[str]]
) -> None:
    md1: MultiDict[str] = any_multidict_class({str(i): str(i) for i in range(100)})
    md2: MultiDict[str] = any_multidict_class({str(i): str(i) for i in range(50, 150)})

    @benchmark
    def _run() -> None:
        assert len(md1.items() | md2.items()) == 150


def test_items_view_sub(
    benchmark: BenchmarkFixture, any_multidict_class: Type[MultiDict[str]]
) -> None:
    md1: MultiDict[str] = any_multidict_class({str(i): str(i) for i in range(100)})
    md2: MultiDict[str] = any_multidict_class({str(i): str(i) for i in range(50, 150)})

    @benchmark
    def _run() -> None:
        assert len(md1.items() - md2.items()) == 50


def test_items_view_xor(
    benchmark: BenchmarkFixture, any_multidict_class: Type[MultiDict[str]]
) -> None:
    md1: MultiDict[str] = any_multidict_class({str(i): str(i) for i in range(100)})
    md2: MultiDict[str] = any_multidict_class({str(i): str(i) for i in range(50, 150)})

    @benchmark
    def _run() -> None:
        assert len(md1.items() ^ md2.items()) == 100


def test_items_view_is_disjoint(
    benchmark: BenchmarkFixture, any_multidict_class: Type[MultiDict[str]]
) -> None:
    md1: MultiDict[str] = any_multidict_class({str(i): str(i) for i in range(100)})
    md2: MultiDict[str] = any_multidict_class({str(i): str(i) for i in range(100, 200)})

    @benchmark
    def _run() -> None:
        assert md1.items().isdisjoint(md2.items())


def test_items_view_repr(
    benchmark: BenchmarkFixture, any_multidict_class: Type[MultiDict[str]]
) -> None:
    md: MultiDict[str] = any_multidict_class({str(i): str(i) for i in range(100)})

    @benchmark
    def _run() -> None:
        repr(md.items())


def test_values_view_repr(
    benchmark: BenchmarkFixture, any_multidict_class: Type[MultiDict[str]]
) -> None:
    md: MultiDict[str] = any_multidict_class({str(i): str(i) for i in range(100)})

    @benchmark
    def _run() -> None:
        repr(md.values())



import pickle
from importlib import import_module
from pathlib import Path
from typing import Union

from multidict import CIMultiDict, MultiDict, istr

TESTS_DIR = Path(__file__).parent.resolve()
_MD_Classes = Union[type[MultiDict[int]], type[CIMultiDict[int]]]


def write(tag: str, cls: _MD_Classes, proto: int) -> None:
    d = cls([("a", 1), ("a", 2)])
    file_basename = f"{cls.__name__.lower()}-{tag}"
    with (TESTS_DIR / f"{file_basename}.pickle.{proto}").open("wb") as f:
        pickle.dump(d, f, proto)


def write_istr(tag: str, cls: type[istr], proto: int) -> None:
    s = cls("str")
    file_basename = f"{cls.__name__.lower()}-{tag}"
    with (TESTS_DIR / f"{file_basename}.pickle.{proto}").open("wb") as f:
        pickle.dump(s, f, proto)


def generate() -> None:
    _impl_map = {
        "c-extension": "_multidict",
        "pure-python": "_multidict_py",
    }
    for proto in range(pickle.HIGHEST_PROTOCOL + 1):
        for tag, impl_name in _impl_map.items():
            impl = import_module(f"multidict.{impl_name}")
            for cls in impl.CIMultiDict, impl.MultiDict:
                write(tag, cls, proto)
            write_istr(tag, impl.istr, proto)


if __name__ == "__main__":
    generate()



import copy
from typing import Union

from multidict import CIMultiDict, CIMultiDictProxy, MultiDict, MultiDictProxy

_MD_Classes = Union[type[MultiDict[int]], type[CIMultiDict[int]]]
_MDP_Classes = Union[type[MultiDictProxy[int]], type[CIMultiDictProxy[int]]]


def test_copy(any_multidict_class: _MD_Classes) -> None:
    d = any_multidict_class()
    d["foo"] = 6
    d2 = d.copy()
    d2["foo"] = 7
    assert d["foo"] == 6
    assert d2["foo"] == 7


def test_copy_proxy(
    any_multidict_class: _MD_Classes, any_multidict_proxy_class: _MDP_Classes
) -> None:
    d = any_multidict_class()
    d["foo"] = 6
    p = any_multidict_proxy_class(d)
    d2 = p.copy()
    d2["foo"] = 7
    assert d["foo"] == 6
    assert p["foo"] == 6
    assert d2["foo"] == 7


def test_copy_std_copy(any_multidict_class: _MD_Classes) -> None:
    d = any_multidict_class()
    d["foo"] = 6
    d2 = copy.copy(d)
    d2["foo"] = 7
    assert d["foo"] == 6
    assert d2["foo"] == 7


def test_ci_multidict_clone(any_multidict_class: _MD_Classes) -> None:
    d = any_multidict_class(foo=6)
    d2 = any_multidict_class(d)
    d2["foo"] = 7
    assert d["foo"] == 6
    assert d2["foo"] == 7



import os
import platform
import sys

from setuptools import Extension, setup

NO_EXTENSIONS = bool(os.environ.get("MULTIDICT_NO_EXTENSIONS"))
DEBUG_BUILD = bool(os.environ.get("MULTIDICT_DEBUG_BUILD"))

if sys.implementation.name != "cpython":
    NO_EXTENSIONS = True

CFLAGS = ["-O0", "-g3", "-UNDEBUG"] if DEBUG_BUILD else ["-O3"]

if platform.system() != "Windows":
    CFLAGS.extend(
        [
            "-std=c99",
            "-Wall",
            "-Wsign-compare",
            "-Wconversion",
            "-fno-strict-aliasing",
            "-pedantic",
        ]
    )

extensions = [
    Extension(
        "multidict._multidict",
        ["multidict/_multidict.c"],
        extra_compile_args=CFLAGS,
    ),
]


if not NO_EXTENSIONS:
    print("*********************")
    print("* Accelerated build *")
    print("*********************")
    setup(ext_modules=extensions)
else:
    print("*********************")
    print("* Pure Python build *")
    print("*********************")
    setup()



from collections.abc import Callable
from typing import TypeVar, Union

import pytest

from multidict import CIMultiDict, CIMultiDictProxy, MultiDict, MultiDictProxy

_T = TypeVar("_T")
_MD_Types = Union[
    MultiDict[_T], CIMultiDict[_T], MultiDictProxy[_T], CIMultiDictProxy[_T]
]
GetVersion = Callable[[_MD_Types[_T]], int]


def test_getversion_bad_param(multidict_getversion_callable: GetVersion[str]) -> None:
    with pytest.raises(TypeError):
        multidict_getversion_callable(1)  # type: ignore[arg-type]


def test_ctor(
    any_multidict_class: type[MultiDict[str]],
    multidict_getversion_callable: GetVersion[str],
) -> None:
    m1 = any_multidict_class()
    v1 = multidict_getversion_callable(m1)
    m2 = any_multidict_class()
    v2 = multidict_getversion_callable(m2)
    assert v1 != v2


def test_add(
    any_multidict_class: type[MultiDict[str]],
    multidict_getversion_callable: GetVersion[str],
) -> None:
    m = any_multidict_class()
    v = multidict_getversion_callable(m)
    m.add("key", "val")
    assert multidict_getversion_callable(m) > v


def test_delitem(
    any_multidict_class: type[MultiDict[str]],
    multidict_getversion_callable: GetVersion[str],
) -> None:
    m = any_multidict_class()
    m.add("key", "val")
    v = multidict_getversion_callable(m)
    del m["key"]
    assert multidict_getversion_callable(m) > v


def test_delitem_not_found(
    any_multidict_class: type[MultiDict[str]],
    multidict_getversion_callable: GetVersion[str],
) -> None:
    m = any_multidict_class()
    m.add("key", "val")
    v = multidict_getversion_callable(m)
    with pytest.raises(KeyError):
        del m["notfound"]
    assert multidict_getversion_callable(m) == v


def test_setitem(
    any_multidict_class: type[MultiDict[str]],
    multidict_getversion_callable: GetVersion[str],
) -> None:
    m = any_multidict_class()
    m.add("key", "val")
    v = multidict_getversion_callable(m)
    m["key"] = "val2"
    assert multidict_getversion_callable(m) > v


def test_setitem_not_found(
    any_multidict_class: type[MultiDict[str]],
    multidict_getversion_callable: GetVersion[str],
) -> None:
    m = any_multidict_class()
    m.add("key", "val")
    v = multidict_getversion_callable(m)
    m["notfound"] = "val2"
    assert multidict_getversion_callable(m) > v


def test_clear(
    any_multidict_class: type[MultiDict[str]],
    multidict_getversion_callable: GetVersion[str],
) -> None:
    m = any_multidict_class()
    m.add("key", "val")
    v = multidict_getversion_callable(m)
    m.clear()
    assert multidict_getversion_callable(m) > v


def test_setdefault(
    any_multidict_class: type[MultiDict[str]],
    multidict_getversion_callable: GetVersion[str],
) -> None:
    m = any_multidict_class()
    m.add("key", "val")
    v = multidict_getversion_callable(m)
    m.setdefault("key2", "val2")
    assert multidict_getversion_callable(m) > v


def test_popone(
    any_multidict_class: type[MultiDict[str]],
    multidict_getversion_callable: GetVersion[str],
) -> None:
    m = any_multidict_class()
    m.add("key", "val")
    v = multidict_getversion_callable(m)
    m.popone("key")
    assert multidict_getversion_callable(m) > v


def test_popone_default(
    any_multidict_class: type[MultiDict[str]],
    multidict_getversion_callable: GetVersion[str],
) -> None:
    m = any_multidict_class()
    m.add("key", "val")
    v = multidict_getversion_callable(m)
    m.popone("key2", "default")
    assert multidict_getversion_callable(m) == v


def test_popone_key_error(
    any_multidict_class: type[MultiDict[str]],
    multidict_getversion_callable: GetVersion[str],
) -> None:
    m = any_multidict_class()
    m.add("key", "val")
    v = multidict_getversion_callable(m)
    with pytest.raises(KeyError):
        m.popone("key2")
    assert multidict_getversion_callable(m) == v


def test_pop(
    any_multidict_class: type[MultiDict[str]],
    multidict_getversion_callable: GetVersion[str],
) -> None:
    m = any_multidict_class()
    m.add("key", "val")
    v = multidict_getversion_callable(m)
    m.pop("key")
    assert multidict_getversion_callable(m) > v


def test_pop_default(
    any_multidict_class: type[MultiDict[str]],
    multidict_getversion_callable: GetVersion[str],
) -> None:
    m = any_multidict_class()
    m.add("key", "val")
    v = multidict_getversion_callable(m)
    m.pop("key2", "default")
    assert multidict_getversion_callable(m) == v


def test_pop_key_error(
    any_multidict_class: type[MultiDict[str]],
    multidict_getversion_callable: GetVersion[str],
) -> None:
    m = any_multidict_class()
    m.add("key", "val")
    v = multidict_getversion_callable(m)
    with pytest.raises(KeyError):
        m.pop("key2")
    assert multidict_getversion_callable(m) == v


def test_popall(
    any_multidict_class: type[MultiDict[str]],
    multidict_getversion_callable: GetVersion[str],
) -> None:
    m = any_multidict_class()
    m.add("key", "val")
    v = multidict_getversion_callable(m)
    m.popall("key")
    assert multidict_getversion_callable(m) > v


def test_popall_default(
    any_multidict_class: type[MultiDict[str]],
    multidict_getversion_callable: GetVersion[str],
) -> None:
    m = any_multidict_class()
    m.add("key", "val")
    v = multidict_getversion_callable(m)
    m.popall("key2", "default")
    assert multidict_getversion_callable(m) == v


def test_popall_key_error(
    any_multidict_class: type[MultiDict[str]],
    multidict_getversion_callable: GetVersion[str],
) -> None:
    m = any_multidict_class()
    m.add("key", "val")
    v = multidict_getversion_callable(m)
    with pytest.raises(KeyError):
        m.popall("key2")
    assert multidict_getversion_callable(m) == v


def test_popitem(
    any_multidict_class: type[MultiDict[str]],
    multidict_getversion_callable: GetVersion[str],
) -> None:
    m = any_multidict_class()
    m.add("key", "val")
    v = multidict_getversion_callable(m)
    m.popitem()
    assert multidict_getversion_callable(m) > v


def test_popitem_key_error(
    any_multidict_class: type[MultiDict[str]],
    multidict_getversion_callable: GetVersion[str],
) -> None:
    m = any_multidict_class()
    v = multidict_getversion_callable(m)
    with pytest.raises(KeyError):
        m.popitem()
    assert multidict_getversion_callable(m) == v



from collections import deque
from typing import Union

from multidict import CIMultiDict, MultiDict

_MD_Classes = Union[type[MultiDict[int]], type[CIMultiDict[int]]]


def test_update_replace(any_multidict_class: _MD_Classes) -> None:
    obj1 = any_multidict_class([("a", 1), ("b", 2), ("a", 3), ("c", 10)])
    obj2 = any_multidict_class([("a", 4), ("b", 5), ("a", 6)])
    obj1.update(obj2)
    expected = [("a", 4), ("b", 5), ("a", 6), ("c", 10)]
    assert list(obj1.items()) == expected


def test_update_append(any_multidict_class: _MD_Classes) -> None:
    obj1 = any_multidict_class([("a", 1), ("b", 2), ("a", 3), ("c", 10)])
    obj2 = any_multidict_class([("a", 4), ("a", 5), ("a", 6)])
    obj1.update(obj2)
    expected = [("a", 4), ("b", 2), ("a", 5), ("c", 10), ("a", 6)]
    assert list(obj1.items()) == expected


def test_update_remove(any_multidict_class: _MD_Classes) -> None:
    obj1 = any_multidict_class([("a", 1), ("b", 2), ("a", 3), ("c", 10)])
    obj2 = any_multidict_class([("a", 4)])
    obj1.update(obj2)
    expected = [("a", 4), ("b", 2), ("c", 10)]
    assert list(obj1.items()) == expected


def test_update_replace_seq(any_multidict_class: _MD_Classes) -> None:
    obj1 = any_multidict_class([("a", 1), ("b", 2), ("a", 3), ("c", 10)])
    obj2 = [("a", 4), ("b", 5), ("a", 6)]
    obj1.update(obj2)
    expected = [("a", 4), ("b", 5), ("a", 6), ("c", 10)]
    assert list(obj1.items()) == expected


def test_update_replace_seq2(any_multidict_class: _MD_Classes) -> None:
    obj1 = any_multidict_class([("a", 1), ("b", 2), ("a", 3), ("c", 10)])
    obj1.update([("a", 4)], b=5, a=6)
    expected = [("a", 4), ("b", 5), ("a", 6), ("c", 10)]
    assert list(obj1.items()) == expected


def test_update_append_seq(any_multidict_class: _MD_Classes) -> None:
    obj1 = any_multidict_class([("a", 1), ("b", 2), ("a", 3), ("c", 10)])
    obj2 = [("a", 4), ("a", 5), ("a", 6)]
    obj1.update(obj2)
    expected = [("a", 4), ("b", 2), ("a", 5), ("c", 10), ("a", 6)]
    assert list(obj1.items()) == expected


def test_update_remove_seq(any_multidict_class: _MD_Classes) -> None:
    obj1 = any_multidict_class([("a", 1), ("b", 2), ("a", 3), ("c", 10)])
    obj2 = [("a", 4)]
    obj1.update(obj2)
    expected = [("a", 4), ("b", 2), ("c", 10)]
    assert list(obj1.items()) == expected


def test_update_md(case_sensitive_multidict_class: type[CIMultiDict[str]]) -> None:
    d = case_sensitive_multidict_class()
    d.add("key", "val1")
    d.add("key", "val2")
    d.add("key2", "val3")

    d.update(key="val")

    assert [("key", "val"), ("key2", "val3")] == list(d.items())


def test_update_istr_ci_md(
    case_insensitive_multidict_class: type[CIMultiDict[str]],
    case_insensitive_str_class: type[str],
) -> None:
    d = case_insensitive_multidict_class()
    d.add(case_insensitive_str_class("KEY"), "val1")
    d.add("key", "val2")
    d.add("key2", "val3")

    d.update({case_insensitive_str_class("key"): "val"})

    assert [("key", "val"), ("key2", "val3")] == list(d.items())


def test_update_ci_md(case_insensitive_multidict_class: type[CIMultiDict[str]]) -> None:
    d = case_insensitive_multidict_class()
    d.add("KEY", "val1")
    d.add("key", "val2")
    d.add("key2", "val3")

    d.update(Key="val")

    assert [("Key", "val"), ("key2", "val3")] == list(d.items())


def test_update_list_arg_and_kwds(any_multidict_class: _MD_Classes) -> None:
    obj = any_multidict_class()
    arg = [("a", 1)]
    obj.update(arg, b=2)
    assert list(obj.items()) == [("a", 1), ("b", 2)]
    assert arg == [("a", 1)]


def test_update_tuple_arg_and_kwds(any_multidict_class: _MD_Classes) -> None:
    obj = any_multidict_class()
    arg = (("a", 1),)
    obj.update(arg, b=2)
    assert list(obj.items()) == [("a", 1), ("b", 2)]
    assert arg == (("a", 1),)


def test_update_deque_arg_and_kwds(any_multidict_class: _MD_Classes) -> None:
    obj = any_multidict_class()
    arg = deque([("a", 1)])
    obj.update(arg, b=2)
    assert list(obj.items()) == [("a", 1), ("b", 2)]
    assert arg == deque([("a", 1)])



"""Test passing invalid arguments to the methods of the MultiDict class."""

from dataclasses import dataclass
from typing import cast

import pytest

from multidict import MultiDict


@dataclass(frozen=True)
class InvalidTestedMethodArgs:
    """A set of arguments passed to methods under test."""

    test_id: str
    positional: tuple[object, ...]
    keyword: dict[str, object]

    def __str__(self) -> str:
        """Render a test identifier as a string."""
        return self.test_id


@pytest.fixture(
    scope="module",
    params=(
        InvalidTestedMethodArgs("no_args", (), {}),
        InvalidTestedMethodArgs("too_many_args", ("a", "b", "c"), {}),
        InvalidTestedMethodArgs("wrong_kwarg", (), {"wrong": 1}),
        InvalidTestedMethodArgs(
            "wrong_kwarg_and_too_many_args",
            ("a",),
            {"wrong": 1},
        ),
    ),
    ids=str,
)
def tested_method_args(
    request: pytest.FixtureRequest,
) -> InvalidTestedMethodArgs:
    """Return an instance of a parameter set."""
    return cast(InvalidTestedMethodArgs, request.param)


@pytest.fixture(scope="module")
def multidict_object(
    any_multidict_class: type[MultiDict[int]],
) -> MultiDict[int]:
    return any_multidict_class([("a", 1), ("a", 2)])


def test_getall_args(
    multidict_object: MultiDict[int],
    tested_method_args: InvalidTestedMethodArgs,
) -> None:
    with pytest.raises(TypeError, match=r".*argument.*"):
        multidict_object.getall(
            *tested_method_args.positional,
            **tested_method_args.keyword,
        )


def test_getone_args(
    multidict_object: MultiDict[int],
    tested_method_args: InvalidTestedMethodArgs,
) -> None:
    with pytest.raises(TypeError, match=r".*argument.*"):
        multidict_object.getone(
            *tested_method_args.positional,
            **tested_method_args.keyword,
        )


def test_get_args(
    multidict_object: MultiDict[int],
    tested_method_args: InvalidTestedMethodArgs,
) -> None:
    with pytest.raises(TypeError, match=r".*argument.*"):
        multidict_object.get(
            *tested_method_args.positional,
            **tested_method_args.keyword,
        )


def test_setdefault_args(
    multidict_object: MultiDict[int],
    tested_method_args: InvalidTestedMethodArgs,
) -> None:
    with pytest.raises(TypeError, match=r".*argument.*"):
        multidict_object.setdefault(
            *tested_method_args.positional,
            **tested_method_args.keyword,
        )


def test_popone_args(
    multidict_object: MultiDict[int],
    tested_method_args: InvalidTestedMethodArgs,
) -> None:
    with pytest.raises(TypeError, match=r".*argument.*"):
        multidict_object.popone(
            *tested_method_args.positional,
            **tested_method_args.keyword,
        )


def test_pop_args(
    multidict_object: MultiDict[int],
    tested_method_args: InvalidTestedMethodArgs,
) -> None:
    with pytest.raises(TypeError, match=r".*argument.*"):
        multidict_object.pop(
            *tested_method_args.positional,
            **tested_method_args.keyword,
        )


def test_popall_args(
    multidict_object: MultiDict[int],
    tested_method_args: InvalidTestedMethodArgs,
) -> None:
    with pytest.raises(TypeError, match=r".*argument.*"):
        multidict_object.popall(
            *tested_method_args.positional,
            **tested_method_args.keyword,
        )



import gc
import sys
from typing import Callable, Type

import pytest

IMPLEMENTATION = getattr(sys, "implementation")  # to suppress mypy error
GIL_ENABLED = getattr(sys, "_is_gil_enabled", lambda: True)()


def test_ctor(case_insensitive_str_class: Type[str]) -> None:
    s = case_insensitive_str_class()
    assert "" == s


def test_ctor_str(case_insensitive_str_class: Type[str]) -> None:
    s = case_insensitive_str_class("aBcD")
    assert "aBcD" == s


def test_ctor_istr(case_insensitive_str_class: Type[str]) -> None:
    s = case_insensitive_str_class("A")
    s2 = case_insensitive_str_class(s)
    assert "A" == s
    assert s == s2


def test_ctor_buffer(case_insensitive_str_class: Type[str]) -> None:
    s = case_insensitive_str_class(b"aBc")
    assert "b'aBc'" == s


def test_ctor_repr(case_insensitive_str_class: Type[str]) -> None:
    s = case_insensitive_str_class(None)
    assert "None" == s


def test_str(case_insensitive_str_class: Type[str]) -> None:
    s = case_insensitive_str_class("aBcD")
    s1 = str(s)
    assert s1 == "aBcD"
    assert type(s1) is str


def test_eq(case_insensitive_str_class: Type[str]) -> None:
    s1 = "Abc"
    s2 = case_insensitive_str_class(s1)
    assert s1 == s2


@pytest.fixture
def create_istrs(case_insensitive_str_class: Type[str]) -> Callable[[], None]:
    """Make a callable populating memory with a few ``istr`` objects."""

    def _create_strs() -> None:
        case_insensitive_str_class("foobarbaz")
        istr2 = case_insensitive_str_class()
        case_insensitive_str_class(istr2)

    return _create_strs


@pytest.mark.skipif(
    IMPLEMENTATION.name != "cpython",
    reason="PyPy has different GC implementation",
)
@pytest.mark.skipif(
    not GIL_ENABLED,
    reason="free threading has different GC implementation",
)
def test_leak(create_istrs: Callable[[], None]) -> None:
    gc.collect()
    cnt = len(gc.get_objects())
    for _ in range(10000):
        create_istrs()

    gc.collect()
    cnt2 = len(gc.get_objects())
    assert abs(cnt - cnt2) < 10  # on other GC impls these numbers are not equal
